#
# CheckCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Check command """

import os
import json
from cati.cmdline.BaseCommand import BaseCommand
from cati.cmdline import pr, ansi, ArgParser
from cati.frontend.RootRequired import require_root_permission
from .StateCommand import StateCommand
from .InstallCommand import InstallCommand
from cati.cmdline.components import ReposListErrorShower
from cati.dotcati.Pkg import Pkg
from cati.helpers.hash import calc_file_sha256
from cati.frontend import Env
from cati.repo.Repo import Repo

class CheckCommand(BaseCommand):
    """ Check command """
    def help(self):
        """
        checks system health and security

        this command checks system health and packages security and static files

        Usage: sudo cati check [options]

        Options:
        -q|--quiet: quiet output
        -v|--verbose: verbose output
        --autofix: automatically fix problems by re-installing some packages
        """
        pass

    def config(self) -> dict:
        """ Define and config this command """
        return {
            'name': 'check',
            'options': {
                '--quiet': [False, False],
                '-q': [False, False],
                '--verbose': [False, False],
                '-v': [False, False],
                '--autofix': [False, False],
            },
            'max_args_count': 0,
            'min_args_count': 0,
        }

    def run(self):
        """ Run command """
        
        # require root permission
        require_root_permission()

        result_code = 0

        packages_to_reinstall = []

        if not self.is_quiet():
            pr.p('Starting checking system health and security...')
            pr.p('===============================================')

        # check state
        state_cmd = StateCommand()
        out = state_cmd.handle(ArgParser.parse(['cati', 'state']))
        if out > 0:
            return out

        # search for conflict and dependency corruptions
        if not self.is_quiet():
            pr.p('Checking dependency and conflict corruptions...')
        dependency_problems = []
        conflict_problems = []
        installed_packages = Pkg.installed_list()['list']
        for pkg in installed_packages:
            if self.is_verbose():
                pr.p('[info] checking dependencies and conflicts for ' + pkg.data['name'] + '...')
            for dp in pkg.get_depends():
                if not Pkg.check_state(dp):
                    dependency_problems.append([
                        pkg, dp
                    ])
            for conflict in pkg.get_conflicts():
                if Pkg.check_state(conflict):
                    conflict_problems.append([
                        pkg, conflict
                    ])

        if dependency_problems or conflict_problems:
            for depend in dependency_problems:
                pr.p(ansi.red + 'dependency problem for ' + depend[0].data['name'] + ': ' + depend[1] + ansi.reset)
                packages_to_reinstall.append(depend[0])
            for conflict in conflict_problems:
                pr.p(ansi.red + 'conflict problem for ' + conflict[0].data['name'] + ': ' + conflict[1] + ansi.reset)
                packages_to_reinstall.append(conflict[0])
            result_code = 1
        else:
            pr.p(ansi.green + 'There is not any conflict or dependnecy problem and everything is ok' + ansi.reset)

        # check static files
        if not self.is_quiet():
            pr.p('Checking packages static files...')
        staticfile_problems = []
        for pkg in installed_packages:
            if self.is_verbose():
                pr.p('[info] checking static files for ' + pkg.data['name'] + '...')
            files = pkg.installed_static_files()
            for f in files:
                f[1] = Env.base_path(f[1])
                if os.path.isfile(f[1]):
                    wanted_hash = f[0]
                    current_hash = calc_file_sha256(f[1])
                    if wanted_hash != current_hash:
                        staticfile_problems.append([
                            pkg,
                            f,
                            'file is changed'
                        ])
                else:
                    staticfile_problems.append([
                        pkg,
                        f,
                        'file is deleted'
                    ])
        if staticfile_problems:
            for problem in staticfile_problems:
                pr.p(ansi.red + 'staticfile problem in package ' + problem[0].data['name'] + ': ' + problem[1][1] + ': ' + problem[2] + ansi.reset)
                packages_to_reinstall.append(problem[0])
            result_code = 1
        else:
            pr.p(ansi.green + 'all of static files are ok' + ansi.reset)
        
        # check repos config files health
        if not self.is_quiet():
            pr.p('Checking cati configuration files...')
        if self.is_verbose():
                pr.p('[info] checking repositories config...')
        repos = Repo.get_list()
        pr.p(ansi.red, end='')
        ReposListErrorShower.show(repos)
        pr.p(ansi.reset, end='')
        is_any_repo_error = False
        for repo in repos:
            if repo.syntax_errors:
                is_any_repo_error = True
                result_code = 1
        if not is_any_repo_error:
            pr.p(ansi.green + 'all of cati configuration files are ok' + ansi.reset)

        # check database files
        if not self.is_quiet():
            pr.p('Checking cati database...')
        database_problems = []
        for f in os.listdir(Env.installed_lists()):
            if self.is_verbose():
                pr.p('[info] checking database install dir for ' + f + '...')
            if not os.path.isfile(Env.installed_lists('/' + f + '/files')) or not os.path.isfile(Env.installed_lists('/' + f + '/ver')):
                database_problems.append('installed packages database: directory ' + Env.installed_lists('/' + f) + ' is corrupt')
        for f in os.listdir(Env.security_blacklist()):
            if self.is_verbose():
                pr.p('[info] checking security blacklist part ' + f + '...')
            if not os.path.isfile(Env.security_blacklist('/' + f)):
                database_problems.append('security blacklist: an directory detected: ' + Env.security_blacklist('/' + f))
            else:
                tmp = open(Env.security_blacklist('/' + f), 'r')
                try:
                    json.loads(tmp.read())
                except:
                    database_problems.append('security blacklist: invalid json data in ' + Env.security_blacklist('/' + f))
        if database_problems:
            for problem in database_problems:
                pr.p(ansi.red + 'database: ' + problem + ansi.reset)
            result_code = 1
        else:
            pr.p(ansi.green + 'all of cati database is ok' + ansi.reset)

        if not self.is_quiet():
            if packages_to_reinstall:
                pr.p(ansi.blue + 'We suggest re-install this packages:')
                for pkg in packages_to_reinstall:
                    pr.p('- ' + pkg.data['name'])
                if not self.has_option('--autofix'):
                    pr.p('use --autofix option to re-install them or do this manually')
                    pr.p(ansi.reset, end='')
                else:
                    pr.p(ansi.reset, end='')
                    packages_names = [pkg.data['name'] for pkg in packages_to_reinstall]
                    install_cmd = InstallCommand()
                    args = ['cati', 'install', '--reinstall', *packages_names]
                    cmd_str = ''
                    for arg in args:
                        cmd_str += arg + ' '
                    cmd_str = cmd_str.strip()
                    pr.p(cmd_str)
                    return install_cmd.handle(ArgParser.parse(args))

        return result_code



#
# ShowCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Show command """

from cati.cmdline.BaseCommand import BaseCommand
from cati.cmdline import pr, ansi
from cati.dotcati.Pkg import Pkg
from cati.cmdline.components import PackageShower

class ShowCommand(BaseCommand):
    """ Show command """
    def help(self):
        """
        shows details of packages

        Usage: cati show pkg1 pkg2 ... [options]

        Options:
        --versions: shows versions list of packages
        """
        pass

    def config(self) -> dict:
        """ Define and config this command """
        return {
            'name': 'show',
            'options': {
                '--versions': [False, False],
            },
            'max_args_count': None,
            'min_args_count': 1,
        }

    def run(self):
        """ Run command """

        pr.p('Loading packages list...')
        pr.p('========================')

        loaded_packages = []

        for argument in self.arguments:
            arg_parts = argument.split('=')
            if len(arg_parts) == 1:
                # load last version as default
                pkg = Pkg.load_last(argument)
            else:
                # load specify version
                pkg = Pkg.load_version(arg_parts[0], arg_parts[1])
                if pkg == 1:
                    pkg = False
                elif pkg == 2:
                    self.message('package "' + arg_parts[0] + '" has not version "' + arg_parts[1] + '"' + ansi.reset, before=ansi.red)
                    continue
            if pkg:
                loaded_packages.append(pkg)
            else:
                self.message('unknow package "' + argument + '"' + ansi.reset, before=ansi.red)

        if not loaded_packages:
            return 1

        # show loaded packages
        for pkg in loaded_packages:
            if self.has_option('--versions'):
                versions_list = pkg.get_versions_list()
                pr.p(pkg.data['name'] + ':')
                for ver in versions_list:
                    pr.p(' ' + ver[0] + ':' + ver[1])
            else:
                PackageShower.show(pkg.data)
            if len(loaded_packages) > 1:
                pr.p('---------------------')



#
# SearchCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Search command """

from cati.cmdline.BaseCommand import BaseCommand
from cati.cmdline import pr, ansi, ArgParser
from .ListCommand import ListCommand

class SearchCommand(BaseCommand):
    """ Search command """
    def help(self):
        """
        search between packages by name and description

        (this command is a alias for `cati list --search='word'`)

        Usage: cati search 'word'
        """
        pass

    def config(self) -> dict:
        """ Define and config this command """
        return {
            'name': 'search',
            'options': {
            },
            'max_args_count': None,
            'min_args_count': 1,
        }

    def run(self):
        """ Run command """

        # join all of arguments as one argument
        full_query_string = ''
        if len(self.arguments) > 1:
            for arg in self.arguments:
                full_query_string += arg + ' '
            full_query_string = full_query_string[:len(full_query_string)-1]
        else:
            full_query_string = self.arguments[0]

        # search arguments with `list --search` command
        arguments = ['--search=' + full_query_string]
        arguments.insert(0, 'cati')
        arguments.insert(1, 'list')
        list_command = ListCommand()
        return list_command.handle(ArgParser.parse(arguments))



#
# Temp.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Handle temp files """

import random
import os
import shutil

created_temp_dirs = []
created_temp_files = []

def make_dir() -> str:
    """ Makes a temp directory and returns path of that directory """
    global created_temp_dirs
    path = '/tmp/' + 'cati-temp-' + str(random.random())

    if os.path.exists(path):
        return make_dir()

    os.mkdir(path)
    created_temp_dirs.append(path)
    return path

def make_file() -> str:
    """ Makes a temp file and returns path of that file """
    global created_temp_files
    path = '/tmp/' + 'cati-temp-' + str(random.random())

    if os.path.exists(path):
        return make_file()

    f = open(path, 'w')
    f.write('')
    f.close()
    created_temp_dirs.append(path)
    return path

def clean():
    """ Clear all of created temp files by program """
    global created_temp_dirs, created_temp_files

    for f in created_temp_files:
        if os.path.isfile(f):
            try:
                os.remove(f)
            except:
                pass

    for d in created_temp_dirs:
        if os.path.isdir(d):
            try:
                shutil.rmtree(d)
            except:
                pass

    created_temp_dirs = []
    created_temp_files = []



#
# Installer.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Dotcati package installer """

import os
import json
import time
import hashlib
from .ArchiveModel import BaseArchive
from cati.frontend import Env, Temp, SysArch, SecurityBlacklist
from . import ListUpdater
from .Pkg import Pkg
from .exceptions import DependencyError, ConflictError, PackageScriptError, PackageIsInSecurityBlacklist, FileConflictError
from cati.transaction.BaseTransaction import BaseTransaction
from cati.helpers.hash import calc_file_sha256, calc_file_sha512, calc_file_md5

class Installer:
    """ Dotcati package installer """

    def load_files(self, path: str, base_temp_path: str):
        """
        Loads list of package files from extracted temp dir

        you should set `self.loaded_files` list before call this.
        this method will put loaded files in self.loaded_files list

        Args:
            path (str): that directory you want to load files list from that
            base_temp_path (str): base path of package extracted temp directory
        """

        for item in os.listdir(path):
            if os.path.isfile(path + '/' + item):
                self.loaded_files.append([(path + '/' + item)[len(base_temp_path):], path + '/' + item])
            else:
                self.loaded_files.append([(path + '/' + item)[len(base_temp_path):], path + '/' + item])
                self.load_files(path + '/' + item, base_temp_path)

    def copy_once_file(self, paths):
        """ Copy one of package files (this method is called from `copy_files` method) """
        if os.path.isfile(paths[1]):
            if paths[0] in self.conffiles:
                self.copied_files.append('cf:' + paths[0])
                old_conffiles = [item[-1] for item in self.old_conffiles]
                if paths[0] in old_conffiles:
                    f_hash = calc_file_sha256(paths[1])
                    if [f_hash, paths[0]] in self.old_conffiles:
                        self.uncopied_conffiles[paths[0]] = f_hash
                        return
                    else:
                        if self.keep_conffiles:
                            self.uncopied_conffiles[paths[0]] = f_hash
                            return
            else:
                self.copied_files.append('f:' + paths[0])
            os.system('cp "' + paths[1] + '" "' + Env.base_path(paths[0]) + '"')
        else:
            os.mkdir(Env.base_path(paths[0]))
            if paths[1] in self.conffiles:
                self.copied_files.append('cd:' + paths[0])
            else:
                self.copied_files.append('d:' + paths[0])

    def copy_files(self, pkg: BaseArchive, directory_not_empty_event, target_path='') -> list:
        """
        Copy package files on system

        Args:
            pkg (BaseArchive): the package archive object
            directory_not_empty_event (callable): the function will be run when we want to delete old direcotry
                                      of package and that is not empty.
            target_path (str): target path prefix of copied files location. default is `/` means
                               copies files on the root directory. you can change it.

        Returns:
            list[str]: list of copied files
        """
        # load package old files list
        old_files = []
        if os.path.isfile(Env.installed_lists('/' + pkg.data['name'] + '/files')):
            try:
                f = open(Env.installed_lists('/' + pkg.data['name'] + '/files'), 'r')
                for line in f.read().strip().split('\n'):
                    if line != '':
                        old_files.append(line.strip())
            except:
                pass
        old_files = list(reversed(old_files))

        # load unremoved conffiles list
        unremoved_conffiles_f = open(Env.unremoved_conffiles(), 'r')
        unremoved_conffiles = unremoved_conffiles_f.read().strip().split('\n')
        unremoved_conffiles_f.close()

        temp_dir = self.extracted_package_dir

        # load files list from `files` directory of package
        self.loaded_files = []
        self.load_files(temp_dir + '/files', temp_dir + '/files')
        self.loaded_files = [[target_path + f[0], f[1]] for f in self.loaded_files]

        # check file conflicts
        all_installed_files = Pkg.get_all_installed_files_list()
        for lf in self.loaded_files:
            if not os.path.isdir(lf[1]):
                for insf in all_installed_files:
                    if insf[0].split(':', 1)[0] != pkg.data['name']:
                        if lf[0] == insf[2]:
                            insf_pkg = Pkg.load_last(insf[0])
                            if insf_pkg:
                                insf[0] = insf[0] + ':' + insf_pkg.installed()
                            # check package is in `replaces` list
                            do_raise_error = True
                            replaces_list = pkg.get_replaces()
                            for rep in replaces_list:
                                if Pkg.check_state(rep):
                                    tmp_parts = rep.split(' ')
                                    tmp_parts = tmp_parts[0].split('>')
                                    tmp_parts = tmp_parts[0].split('<')
                                    tmp_parts = tmp_parts[0].split('=')
                                    tmp_parts = tmp_parts[0]
                                    if tmp_parts == insf[0].split(':', 1)[0]:
                                        do_raise_error = False
                            if do_raise_error:
                                raise FileConflictError(
                                    'package ' + pkg.data['name'] + ':' + pkg.data['version'] + ' and ' + insf[0] + ' both has file "' + lf[0] + '"'
                                )

        # copy loaded files
        self.copied_files = []
        for f in self.loaded_files:
            if os.path.exists(Env.base_path(f[0])):
                if os.path.isfile(Env.base_path(f[0])):
                    if ('f:' + f[0]) in old_files or ('cf:' + f[0]) in old_files:
                        self.copy_once_file(f)
                        try:
                            old_files.pop(old_files.index(('f:' + f[0])))
                        except:
                            pass
                    else:
                        if f[0] in unremoved_conffiles:
                            self.copy_once_file(f)
                            unremoved_conffiles.pop(unremoved_conffiles.index(f[0]))
                        else:
                            self.copy_once_file(f)
                else:
                    if ('d:' + f[0]) in old_files or ('cd:' + f[0]) in old_files:
                        if ('cd:' + f[0]) in old_files:
                            self.copied_files.append('cd:' + f[0])
                            old_files.pop(old_files.index(('cd:' + f[0])))
                        else:
                            self.copied_files.append('d:' + f[0])
                            old_files.pop(old_files.index(('d:' + f[0])))
                    else:
                        if f[0] in unremoved_conffiles:
                            self.copied_files.append('d:' + f[0])
                            unremoved_conffiles.pop(unremoved_conffiles.index(f[0]))
            else:
                self.copy_once_file(f)

        # delete not wanted old files
        for item in old_files:
            parts = item.strip().split(':', 1)
            if parts[0] == 'cf' or parts[0] == 'cd':
                pass
            else:
                if os.path.isfile(parts[1]):
                    os.remove(parts[1])
                else:
                    try:
                        os.rmdir(parts[1])
                    except:
                        # directory is not emptyr
                        directory_not_empty_event(pkg, parts[1])

        # write new unremoved conffiles list
        unremoved_conffiles_f = open(Env.unremoved_conffiles(), 'w')
        new_content = ''
        for item in unremoved_conffiles:
            new_content += item + '\n'
        unremoved_conffiles_f.write(new_content)
        unremoved_conffiles_f.close()

        return self.copied_files

    def check_dep_and_conf(self, pkg: BaseArchive):
        """
        Checks package dependencies and conflicts.

        Args:
            pkg: (BaseArchive) the package archive object

        Raises:
            DependencyError: when a dependency is not installed (dependency will put in exception message)
            ConflictError: when a conflict is installed (conflict will put in exception message)
        """

        # load package dependencies
        try:
            depends = pkg.data['depends']
        except:
            depends = []

        # load package conflicts
        try:
            conflicts = pkg.data['conflicts']
        except:
            conflicts = []

        # load package reverse cconflicts (packages has conflict to this package)
        reverse_conflicts = pkg.get_reverse_conflicts()

        for dep in depends:
            if not Pkg.check_state(dep):
                raise DependencyError(dep)

        for conflict in conflicts:
            if Pkg.check_state(conflict):
                raise ConflictError(conflict)

        # check reverse conflicts
        if reverse_conflicts:
            raise ConflictError('reverse conflict with ' + reverse_conflicts[0].data['name'] + ': ' + reverse_conflicts[0].conflict_error)

    def run_script(self, script_name: str, script_path=None):
        """
        runs an script in the package (only install scripts)

        Args:
            script_name (str): name of that script you want to run
            script_path (str, None): you can change the default path of script file with this argument

        Raises:
            dotcati.exceptions.PackageScriptError: when script returns non-zero exit code
                (exit code will put in `error_code` attribute of exception object)
        """
        if script_path == None:
            script_path = self.extracted_package_dir + '/scripts/' + script_name
        if os.path.isfile(script_path):
            # script exists, run script
            os.system('chmod +x "' + script_path + '"')
            # run script and pass old version of package (if currently installed)
            old_version = ''
            try:
                old_version = self.old_version
            except:
                if Pkg.is_installed(self.pkg.data['name']):
                    self.old_version = '"' + Pkg.installed_version(self.pkg.data['name']) + '"'
                    old_version = self.old_version
            result = os.system(script_path + ' ' + old_version)
            if result != 0:
                tmp = PackageScriptError("script " + script_name + ' returned non-zero code ' + str(result))
                tmp.error_code = result
                raise tmp

    def check_security_blacklist(self, pkg: BaseArchive):
        """
        checks package sha256, sha512 and md5 and checks this hashes in security blacklist.

        Args:
            pkg (BaseArchive): the package archive object

        Raises:
            dotcati.exceptions.PackageIsInSecurityBlacklist: when package is blocked in security blacklist
                (the blacklist details will put in `blacklist_item` attribute of exception object)
        """
        # calculate hashes
        sha256 = calc_file_sha256(pkg.package_file_path)
        sha512 = calc_file_sha512(pkg.package_file_path)
        md5 = calc_file_md5(pkg.package_file_path)

        # check blacklist
        blacklist = SecurityBlacklist.get_list()
        for item in blacklist:
            if item['md5'] == md5 and item['sha512'] == sha512 and item['sha256'] == sha256:
                ex = PackageIsInSecurityBlacklist()
                ex.blacklist_item = item
                raise ex

    def install(self, pkg: BaseArchive, index_updater_events: dict, installer_events: dict, is_manual=True, run_scripts=True, target_path='', keep_conffiles=False, check_security_blacklist=True):
        """
        Install .cati package

        Args:
            pkg (BaseArchive): the package archive object
            index_updater_events (dict): events will be passed to `dotcati.ListUpdater.update_indexes()`
            installer_events (dict): The events
                - package_currently_install: gets the current installed version
                - package_new_installs: gets package archive
                - package_installed: will call after package installation
                - dep_and_conflict_error: will run when there is depends or conflict error
                - arch_error: will run when package arch is not sync with sys arch
            is_manual (bool): installed package as manual or not (default is True means manual)
            run_scripts (bool): run package install scripts or not (default is True)
            target_path (str): where is the target path for installed files (will pass to `self.copy_files()`)
            keep_conffiles (bool): stil keep config files if changed (default is True)
            check_security_blacklist (bool): check package is in security blacklist or not
        """

        self.conffiles = pkg.get_conffiles()
        self.pkg = pkg
        self.keep_conffiles = keep_conffiles
        self.uncopied_conffiles = {}

        # check package is in security blacklist
        if check_security_blacklist:
            self.check_security_blacklist(pkg)

        # check package architecture
        if not pkg.data['arch'] in SysArch.allowed_archs():
            return installer_events['arch_error'](pkg)

        # add package data to lists
        self.add_package_to_lists(pkg, index_updater_events)

        # check package dependencies and conflicts
        try:
            self.check_dep_and_conf(pkg)
        except DependencyError as ex:
            return installer_events['dep_and_conflict_error'](pkg, ex)
        except ConflictError as ex:
            return installer_events['dep_and_conflict_error'](pkg, ex)

        # load old conffiles
        self.old_conffiles = []
        try:
            f = open(Env.installed_lists('/' + pkg.data['name'] + '/conffiles'), 'r')
            content = f.read()
            f.close()
            tmp = content.strip().split('\n')
            self.old_conffiles = [item.strip().split('@') for item in tmp]
        except:
            pass

        # extract package in a temp place
        temp_dir = Temp.make_dir()
        os.rmdir(temp_dir)
        try:
            pkg.extractall(temp_dir)
        except IsADirectoryError:
            pass
        self.extracted_package_dir = temp_dir

        # install package
        if Pkg.is_installed(pkg.data['name']):
            installer_events['package_currently_installed'](pkg, Pkg.installed_version(pkg.data['name']))
        else:
            installer_events['package_new_installs'](pkg)

        if run_scripts:
            self.run_script('ins-before')

        copied_files = self.copy_files(pkg, installer_events['directory_not_empty'], target_path)

        # set install configuration
        if not os.path.isdir(Env.installed_lists('/' + pkg.data['name'])):
            os.mkdir(Env.installed_lists('/' + pkg.data['name']))
        f_ver = open(Env.installed_lists('/' + pkg.data['name'] + '/ver'), 'w')
        f_ver.write(pkg.data['version']) # write installed version
        f_ver.close()

        # write copied files list
        f_files = open(Env.installed_lists('/' + pkg.data['name'] + '/files'), 'w')
        copied_files_str = ''
        for copied_file in copied_files:
            copied_files_str += copied_file + '\n'
        f_files.write(copied_files_str.strip()) # write copied files
        f_files.close()

        # write conffiles list
        f_conffiles = open(Env.installed_lists('/' + pkg.data['name'] + '/conffiles'), 'w')
        copied_conffiles_str = ''
        for copied_conffile in copied_files:
            if copied_conffile.split(':')[0] == 'cf':
                try:
                    conffile_hash = self.uncopied_conffiles[copied_conffile.split(':', 1)[-1]]
                except:
                    conffile_hash = calc_file_sha256(Env.base_path(copied_conffile.split(':', 1)[-1]))
                copied_conffiles_str += conffile_hash + '@' + copied_conffile.split(':', 1)[-1] + '\n'
        f_conffiles.write(copied_conffiles_str.strip()) # write copied conffiles
        f_conffiles.close()

        # copy `any` script
        if os.path.isfile(self.extracted_package_dir + '/scripts/any'):
            os.system('cp "' + self.extracted_package_dir + '/scripts/any' + '" "' + Env.any_scripts('/' + pkg.data['name']) + '"')

        # save static files list
        static_files_list = pkg.get_static_files()
        f_static_files = open(Env.installed_lists('/' + pkg.data['name'] + '/staticfiles'), 'w')
        static_files_str = ''
        for copied_file in copied_files:
            copied_file_path = copied_file.split(':', 1)[1]
            if copied_file_path in static_files_list:
                if os.path.isfile(Env.base_path('/' + copied_file_path)):
                    # calculate file sha256 sum
                    copied_file_sha256 = calc_file_sha256(Env.base_path('/' + copied_file_path))
                    # add file to list
                    static_files_str += copied_file_sha256 + '@' + copied_file_path + '\n'
        f_static_files.write(static_files_str.strip()) # write copied files
        f_static_files.close()

        f_installed_at = open(Env.installed_lists('/' + pkg.data['name'] + '/installed_at'), 'w')
        f_installed_at.write(str(time.time())) # write time (installed at)
        f_installed_at.close()

        if is_manual:
            f_manual = open(Env.installed_lists('/' + pkg.data['name'] + '/manual'), 'w')
            f_manual.write('')
            f_manual.close()

        if run_scripts:
            self.run_script('ins-after')

        # copy remove scripts
        if os.path.isfile(self.extracted_package_dir + '/scripts/rm-before'):
            os.system(
                'cp "' + self.extracted_package_dir + '/scripts/rm-before' + '" "' + Env.installed_lists('/' + pkg.data['name'] + '/rm-before') + '"'
            )
        if os.path.isfile(self.extracted_package_dir + '/scripts/rm-after'):
            os.system(
                'cp "' + self.extracted_package_dir + '/scripts/rm-after' + '" "' + Env.installed_lists('/' + pkg.data['name'] + '/rm-after') + '"'
            )

        # pop package from state
        BaseTransaction.pop_state()

        # call package installed event
        installer_events['package_installed'](pkg)

    def add_package_to_lists(self, pkg: Pkg, index_updater_events: dict):
        """
        Adds the package information to database
        
        Args:
            pkg (Pkg): the package
            index_updater_events (dict): events for index updater
        """
        if not os.path.isdir(Env.packages_lists('/' + pkg.data['name'])):
            os.mkdir(Env.packages_lists('/' + pkg.data['name']))
        lists_path = Env.packages_lists('/' + pkg.data['name'] + '/' + pkg.data['version'] + '-' + pkg.data['arch'])
        try:
            lists_f = open(lists_path, 'r')
            old_repo = json.loads(lists_f.read())['repo']
            lists_f.close()
        except:
            old_repo = 'Local'
        try:
            lists_f = open(lists_path, 'r')
            old_file_path = json.loads(lists_f.read())['file_path']
            lists_f.close()
        except:
            old_file_path = False
        try:
            lists_f = open(lists_path, 'r')
            old_file_sha256 = json.loads(lists_f.read())['file_sha256']
            lists_f.close()
        except:
            old_file_sha256 = False
        try:
            lists_f = open(lists_path, 'r')
            old_file_md5 = json.loads(lists_f.read())['file_md5']
            lists_f.close()
        except:
            old_file_md5 = False
        lists_f = open(lists_path, 'w')
        pkg.data['repo'] = old_repo
        if old_file_path != False:
            pkg.data['file_path'] = old_file_path
        tmp_pkg_data = pkg.data
        if old_file_md5:
            tmp_pkg_data['file_md5'] = old_file_md5
        if old_file_sha256:
            tmp_pkg_data['file_sha256'] = old_file_sha256
        tmp_pkg_data['files'] = ['/' + member[6:] for member in pkg.members() if member[:6] == 'files/']
        lists_f.write(json.dumps(tmp_pkg_data))
        lists_f.close()
        ListUpdater.update_indexes(index_updater_events)
        ListUpdater.index_reverse_depends_and_conflicts(pkg)



#
# hash.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Some helper functions for hashing """

import hashlib

def calc_file_sha256(filepath: str) -> str:
    """
    gets filepath and calculates sha256 sum of that
    
    Args:
        filepath (str): filepath you want to calculate hash of that

    Returns:
        str: calculated hash
    """
    sha256_hash = hashlib.sha256()
    f = open(filepath, 'rb')
    for byte_block in iter(lambda: f.read(4096),b""):
        sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()

def calc_file_sha512(filepath):
    """
    gets filepath and calculates sha512 sum of that
    
    Args:
        filepath (str): filepath you want to calculate hash of that

    Returns:
        str: calculated hash
    """
    sha512_hash = hashlib.sha512()
    f = open(filepath, 'rb')
    for byte_block in iter(lambda: f.read(4096),b""):
        sha512_hash.update(byte_block)
    return sha512_hash.hexdigest()

def calc_file_md5(filepath):
    """
    gets filepath and calculates md5 sum of that
    
    Args:
        filepath (str): filepath you want to calculate hash of that

    Returns:
        str: calculated hash
    """
    md5_hash = hashlib.md5()
    f = open(filepath, 'rb')
    for byte_block in iter(lambda: f.read(4096),b""):
        md5_hash.update(byte_block)
    return md5_hash.hexdigest()



#
# PkgConvertor.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Deb and Rpm to cati package convertor """

import os
import sys
import shutil
import json
from cati.frontend import Temp
from cati.cmdline.commands import PkgCommand
from cati.cmdline import ArgParser

def convert_depends_list(debian_depends_control_value: str) -> list:
    """
    Converts the debian control file `depends/conflicts` items syntax to cati depends/conflicts list

    Args:
        debian_depends_control_value (str): value of package debian control field

    Returns:
        list: the list of depends/conflicts (converted to cati strcture)
    """
    depends_list = debian_depends_control_value.strip().split(',')
    depends_list = [item.strip().split('|') for item in depends_list]
    cati_data_json_list = []

    for item in depends_list:
        current_item = ''
        for part in item:
            part = part.strip()
            part_parts = part.split(' ', 1)
            pkg_name = part_parts[0]
            pkg_name = pkg_name.split(':')[0]
            version_part = None
            if len(part_parts) > 1:
                version_part = part_parts[1]
                version_part = version_part.strip().strip('(').strip(')')
                version_part = version_part.replace('>>', '>')
                version_part = version_part.replace('<<', '<')
            current_item += pkg_name + ' '
            if version_part:
                current_item += str(version_part)
            current_item += '| '
        current_item = current_item.strip()
        current_item = current_item[:len(current_item)-1]
        current_item = current_item.strip()
        cati_data_json_list.append(current_item)

    return cati_data_json_list

def deb2cati(file_path: str) -> str:
    """
    Converts deb package to cati package and returns generated cati package file path

    Args:
        file_path: deb package filepath

    Returns:
        returns generated cati package filepath
    """
    file_path = os.path.abspath(file_path)
    tmp_dir = Temp.make_dir()
    cwd = os.getcwd()
    os.chdir(tmp_dir)
    try:
        shutil.copy(file_path, './package.deb')

        script = '''
        ar -x package.deb
        mkdir cati
        mkdir cati/control cati/scripts
        mv control.tar.* cati/control
        mkdir cati/files
        mv data.tar.* cati/files
        rm debian-binary package.deb
        touch cati/data.json
        '''.strip().split('\n')
        script = [l.strip() for l in script]

        for line in script:
            result = os.system(line)
            if result != 0:
                # script has error, return current filepath
                os.chdir(cwd)
                return file_path
        os.chdir('cati/control')
        os.system('tar -xf control.tar.*')
        os.system('rm control.tar.*')
        os.chdir('..')
        os.chdir('files')
        os.system('tar -xf data.tar.*')
        os.system('rm data.tar.*')
        os.chdir('..')

        # convert content data to cati json
        control_f = open('control/control', 'r').read()
        control_f_lines = control_f.strip().split('\n')
        control_fields = {}
        tmp_last_key = None
        for line in control_f_lines:
            if line != '':
                if line[0] == ' ':
                    if tmp_last_key != None:
                        control_fields[tmp_last_key] += '\n' + line
                else:
                    key = line.split(':', 1)[0]
                    value = line.split(':', 1)[1].strip()
                    tmp_last_key = key
                    control_fields[key] = value

        # convert scripts
        if os.path.isfile('control/preinst'):
            shutil.copy('control/preinst', 'scripts/ins-before')

        if os.path.isfile('control/postinst'):
            shutil.copy('control/postinst', 'scripts/ins-after')

        if os.path.isfile('control/prerm'):
            shutil.copy('control/prerm', 'scripts/rm-before')

        if os.path.isfile('control/postrm'):
            shutil.copy('control/postrm', 'scripts/rm-after')

        # convert control fields to cati data.json
        cati_data = {}
        for k in control_fields:
            if k == 'Package':
                cati_data['name'] = control_fields[k].strip()
            elif k == 'Version':
                cati_data['version'] = control_fields[k].strip()
            elif k == 'Architecture':
                cati_data['arch'] = control_fields[k].strip()
            elif k == 'Maintainer':
                cati_data['maintainer'] = control_fields[k].strip()
            elif k == 'Original-Maintainer':
                cati_data['X-Original-Maintainer'] = control_fields[k].strip()
            elif k == 'Uploaders':
                cati_data['uploaders'] = control_fields[k].strip().split(',')
                cati_data['uploaders'] = [a.strip() for a in cati_data['uploaders']]
            elif k == 'Description':
                cati_data['description'] = control_fields[k]
            elif k == 'Changed-By':
                cati_data['changed-by'] = control_fields[k]
            elif k == 'Changes':
                cati_data['changes'] = control_fields[k]
            elif k == 'Date':
                cati_data['date'] = control_fields[k]
            elif k == 'Urgency':
                cati_data['urgency'] = control_fields[k]
            elif k == 'Essential':
                cati_data['essential'] = control_fields[k]
                if cati_data['essential'] == 'yes' or cati_data['essential'] == 'Yes':
                    cati_data['essential'] = True
                else:
                    cati_data['essential'] = False
            elif k == 'Homepage':
                cati_data['homepage'] = control_fields[k].strip()
            elif k == 'Section':
                cati_data['category'] = [control_fields[k].strip()]
            elif k == 'Depends' or k == 'Pre-Depends':
                try:
                    cati_data['depends']
                except:
                    cati_data['depends'] = []
                cati_data['depends'] = [*cati_data['depends'], *convert_depends_list(control_fields[k].strip())]
            elif k == 'Conflicts' or k == 'Breaks':
                try:
                    cati_data['conflicts']
                except:
                    cati_data['conflicts'] = []
                cati_data['conflicts'] = [*cati_data['conflicts'], *convert_depends_list(control_fields[k].strip())]
            elif k == 'Recommends':
                cati_data['recommends'] = convert_depends_list(control_fields[k].strip())
            elif k == 'Replaces':
                cati_data['replaces'] = convert_depends_list(control_fields[k].strip())
            elif k == 'Suggests':
                cati_data['suggests'] = convert_depends_list(control_fields[k].strip())
                cati_data['suggests'] = [item_tmp.split(' ')[0] for item_tmp in cati_data['suggests']]
            elif k == 'Enhances':
                cati_data['enhances'] = convert_depends_list(control_fields[k].strip())
                cati_data['enhances'] = [item_tmp.split(' ')[0] for item_tmp in cati_data['enhances']]
            elif k == 'Provides':
                cati_data['provides'] = convert_depends_list(control_fields[k].strip())
                cati_data['provides'] = [item_tmp.split(' ')[0] for item_tmp in cati_data['provides']]
            elif k[0] == 'X' or k[0] == 'x':
                cati_data[k] = control_fields[k].strip()
        os.system('rm control -rf')
        cati_data_f = open('data.json', 'w')
        cati_data_f.write(json.dumps(cati_data))
        cati_data_f.close()
        os.chdir('..')

        # build pkg
        pkg_command = PkgCommand.PkgCommand()
        pkg_command.handle(ArgParser.parse(['cati', 'pkg', 'build', 'cati', '-q']))

        if os.path.isfile('cati.cati'):
            try:
                shutil.copy('cati.cati', file_path + '.cati')
                os.chdir(cwd)
                return file_path + '.cati'
            except:
                tmp_file_path = Temp.make_file()
                shutil.copy('cati.cati', tmp_file_path)
                return tmp_file_path
    except Exception as ex:
        print('error: ' + str(ex))
        os.chdir(cwd)
        return file_path

    os.chdir(cwd)
    return file_path

def rpm2cati(file_path: str) -> str:
    """
    Converts rpm package to cati package and returns generated cati package file path

    Args:
        file_path: rpm package filepath

    Returns:
        returns generated cati package filepath
    """
    # require alien
    if os.system('alien 2> /dev/null') != 256:
        return file_path

    file_path = os.path.abspath(file_path)
    cwd = os.getcwd()
    tmp_dir = Temp.make_dir()
    os.chdir(tmp_dir)

    try:
        shutil.copy(file_path, 'package.rpm')
        res = os.system('alien package.rpm 2> /dev/null')
        if res != 0:
            return file_path

        os.system('mv *.deb package.deb')

        rpm_path = file_path + '.deb'
        try:
            shutil.copy('package.deb', rpm_path)
        except:
            rpm_path = Temp.make_file()
            shutil.copy('package.deb', rpm_path)

        os.chdir(cwd)
        return deb2cati(rpm_path)
    except:
        os.chdir(cwd)
        return file_path



#
# TransactionShower.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" CLI transactions list shower """

from cati.transaction.Calculator import Calculator
from cati.cmdline import ansi, pr

def show(calc: Calculator):
    """
    shows transactions from `Calculator` object.

    Args:
        calc: `transaction.Calculator.Calculator` object
    """
    if calc.to_remove:
        pr.p('The following packages will be removed:')
        for pkg in calc.to_remove:
            pr.p('- ' + ansi.red + pkg.data['name'] + ansi.reset)

    to_install = []
    to_upgrade = []
    to_downgrade = []
    to_reinstall = []
    for pkg in calc.to_install:
        if not pkg.installed():
            to_install.append(pkg)
        elif pkg.compare_version(pkg.wanted_version, pkg.installed()) == 1:
            to_upgrade.append(pkg)
        elif pkg.compare_version(pkg.wanted_version, pkg.installed()) == -1:
            to_downgrade.append(pkg)
        else:
            to_reinstall.append(pkg)
    if to_upgrade:
        pr.p('The following packages will be upgraded:')
        for pkg in to_upgrade:
            pr.p('- ' + ansi.yellow + pkg.data['name'] + '(' + pkg.installed() + ' -> ' + pkg.wanted_version + ') ' + pkg.get_file_size_str() + ansi.reset)

    if to_downgrade:
        pr.p('The following packages will be downgraded:')
        for pkg in to_downgrade:
            pr.p('- ' + ansi.yellow + pkg.data['name'] + '(' + pkg.installed() + ' >> ' + pkg.wanted_version + ') ' + pkg.get_file_size_str() + ansi.reset)

    if to_reinstall:
        pr.p('The following packages will be re-installed:')
        for pkg in to_reinstall:
            pr.p('- ' + ansi.yellow + pkg.data['name'] + '(' + pkg.wanted_version + ') ' + pkg.get_file_size_str() + ansi.reset)

    if to_install:
        pr.p('The following packages will be installed:')
        for pkg in to_install:
            pr.p('- ' + ansi.green + pkg.data['name'] + '(' + pkg.wanted_version + ') ' + pkg.get_file_size_str() + ansi.reset)

    total_download_size = calc.get_total_download_size()

    if total_download_size > 0:
        pr.p('Total download size: ' + str(total_download_size))



#
# ForgetCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Forget command """

import os
import shutil
import glob
from cati.cmdline.BaseCommand import BaseCommand
from cati.cmdline import pr, ansi
from cati.frontend import Env, RootRequired
from cati.dotcati import ListUpdater
from cati.dotcati.Pkg import Pkg

class ForgetCommand(BaseCommand):
    """ Forget command """
    def help(self):
        """
        forgets packages from packages list

        Usage: cati forget pkg1 pkg2 pkg3=1.6.7 ...
        """
        pass

    def config(self) -> dict:
        """ Define and config this command """
        return {
            'name': 'forget',
            'options': {
            },
            'max_args_count': None,
            'min_args_count': 1,
        }

    def empty_method_for_event(a=None, b=None):
        """ an empty method """
        pass

    def run(self):
        """ Run command """

        RootRequired.require_root_permission()

        pr.p('Loading packages list...')
        pr.p('========================')

        loaded_packages = []

        for argument in self.arguments:
            arg_parts = argument.split('=')
            if len(arg_parts) == 1:
                # load last version as default
                pkg = Pkg.load_last(argument)
            else:
                # load specify version
                pkg = Pkg.load_version(arg_parts[0], arg_parts[1])
                if pkg == 1:
                    pkg = False
                elif pkg == 2:
                    self.message('package "' + arg_parts[0] + '" has not version "' + arg_parts[1] + '"' + ansi.reset, before=ansi.red)
                    continue
                else:
                    pkg.only_specify_version = True
            if pkg:
                try:
                    pkg.only_specify_version
                except:
                    pkg.only_specify_version = False
                if pkg.installed():
                    if not pkg.only_specify_version:
                        self.message('package "' + argument + '" is installed. cannot forget installed packages' + ansi.reset, before=ansi.red)
                        continue
                    else:
                        if pkg.installed() == pkg.data['version']:
                            self.message('package ' + argument + ' (' + pkg.data['version'] + ') is installed. cannot forget installed packages' + ansi.reset, before=ansi.red)
                            continue
                loaded_packages.append(pkg)
            else:
                self.message('unknow package "' + argument + '"' + ansi.reset, before=ansi.red)

        if not loaded_packages:
            return 1

        # forget loaded packages
        for pkg in loaded_packages:
            if not pkg.only_specify_version:
                # forget all of package versions
                shutil.rmtree(Env.packages_lists('/' + pkg.data['name']))
                pr.p('Package ' + pkg.data['name'] + ' was forgoten successfully')
            else:
                files = glob.glob(Env.packages_lists('/' + pkg.data['name'] + '/' + pkg.data['version'] + '-*'))
                for f in files:
                    if not '-' in f[len(Env.packages_lists('/' + pkg.data['name'] + '/' + pkg.data['version'] + '-')):]:
                        os.remove(f)
                pr.p('Version ' + pkg.data['version'] + ' of package ' + pkg.data['name'] + ' was forgoten successfully')
            try:
                if len(os.listdir(Env.packages_lists('/' + pkg.data['name']))) <= 1:
                    shutil.rmtree(Env.packages_lists('/' + pkg.data['name']))
            except:
                pass

        ListUpdater.update_indexes({
            'cannot_read_file': self.empty_method_for_event,
            'invalid_json_data': self.empty_method_for_event,
        })



#
# PkgCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Pkg command to work with .cati archives """

import os
from cati.cmdline.BaseCommand import BaseCommand
from cati.cmdline import pr, ansi
from cati.dotcati.Builder import Builder
from cati.dotcati.Installer import Installer
from cati.dotcati.exceptions import InvalidPackageDirException, InvalidPackageFileException, DependencyError, ConflictError, PackageScriptError, PackageIsInSecurityBlacklist, FileConflictError, CannotReadFileException
from cati.dotcati.ArchiveModel import archive_factory, BaseArchive
from cati.frontend.RootRequired import require_root_permission
from cati.frontend import Env
from cati.cmdline.components import PackageShower, StateContentShower
from cati.transaction.BaseTransaction import BaseTransaction

class PkgCommand(BaseCommand):
    """ Pkg command to work with .cati archives """
    def help(self):
        """
        work with .cati packages

        Usage: cati pkg [subcommand] [args] [options]

        Subcommands:
        - build:      build .cati package from directory(s)
        - show:       show content of .cati package(s). options: --files: show package files
        - install:    install a .cati package on system

        Options:
        - install subcommand:
        * --without-scripts: do not run package scripts in installation process
        * --target=[files-install-location-prefix-path]: set files installation prefix
        * --keep-conffiles: don't overwrite new version of config files
        * --force|-f: force install that packages are blocked in securiy blacklist
        """
        pass

    def config(self) -> dict:
        """ Define and config this command """
        return {
            'name': 'pkg',
            'options': {
                '--quiet': [False, False],
                '-q': [False, False],
                '--output': [False, True],
                '-o': [False, True],
                '--files': [False, False],
                '-f': [False, False],
                '--auto': [False, False],
                '--without-scripts': [False, False],
                '--target': [False, True],
                '--dont-ignore-state': [False, False],
                '--keep-conffiles': [False, False],
                '--force': [False, False],
                '-f': [False, False],
            },
            'max_args_count': None,
            'min_args_count': None,
        }

    def sub_build(self):
        """ build subcommand (cati pkg build) """
        if len(self.arguments) <= 1:
            self.message('argument package directory(s) required')
            return 1

        i = 1
        while i < len(self.arguments):
            try:
                output = self.option_value('--output')
                if output == None:
                    output = self.option_value('-o')
                builder = Builder()
                output_package = builder.build(self.arguments[i], output)
                if not self.is_quiet():
                    pr.p(ansi.green + 'Package ' + self.arguments[i] + ' created successfuly in ' + output_package + ansi.reset)
            except FileNotFoundError as ex:
                self.message('directory "' + self.arguments[i] + '" not found' + ansi.reset, before=ansi.red)
                return 1
            except InvalidPackageDirException as ex:
                self.message('cannot build "' + self.arguments[i] + '": ' + str(ex) + ansi.reset, before=ansi.red)
                return 1

            i += 1

    def show_once(self, pkg: BaseArchive):
        """
        shows once package (called from `show` subcommand function)
        gives package data to cli package sohwer component to show package info
        """
        PackageShower.show(pkg.data)
        if self.has_option('--files') or self.has_option('-f'):
            pr.p('Files:')
            pkg.tar.list()
        if len(self.arguments) > 2:
            pr.p('='*50)

    def sub_show(self):
        """ show subcommand (cati pkg show) """
        if len(self.arguments) <= 1:
            self.message('argument package file(s) required')
            return 1

        i = 1
        while i < len(self.arguments):
            try:
                pkg = archive_factory(self.arguments[i], 'r')
                pkg.read()
                self.show_once(pkg)
                pkg.close()
            except FileNotFoundError as ex:
                self.message('file "' + self.arguments[i] + '" not found' + ansi.reset, before=ansi.red)
            except:
                self.message('cannot open "' + self.arguments[i] + '": file is corrupt' + ansi.reset, before=ansi.red)

            i += 1

    def cannot_read_file_event(self, path):
        """
        index updater cannot_read_file event
        will pass to installer and installer passes this to index updater
        """
        self.message('error while reading file "' + path + '". ignored...' + ansi.reset, before=ansi.red)

    def invalid_json_data_event(self, path, content):
        """
        index updater invalid_json_data event
        will pass to installer and installer passes this to index updater
        """
        self.message('invalid json data in "' + path + '". ignored...' + ansi.reset, before=ansi.red)

    def package_currently_installed_event(self, package: BaseArchive, current_version: str):
        """
        installer package_currently_installed event
        will run when package already installed
        gets package and current installed version
        """
        pr.p('Installing ' + ansi.yellow + package.data['name'] + ':' + package.data['version'] + ansi.reset + ' (over ' + ansi.yellow + current_version + ansi.reset + ')...', end=' ')

    def package_new_installs_event(self, package: BaseArchive):
        """
        installer package_new_installs event
        will run when package will NEW installed
        """
        pr.p('Installing ' + ansi.yellow + package.data['name'] + ':' + package.data['version'] + ansi.reset + '...', end=' ')

    def package_installed_event(self, package: BaseArchive):
        """
        installer package_installed event
        will run when package installation finshed
        """
        pr.p(ansi.green + 'OK' + ansi.reset)

    def directory_not_empty_event(self, package: BaseArchive, dirpath: str):
        """
        installer directory_not_empty event
        will run when package old directory is not empty
        """
        pr.e(ansi.yellow + 'warning: directory "' + dirpath + '" is not empty and will not be deleted' + ansi.reset)

    def dep_and_conflict_error_event(self, pkg: BaseArchive, ex: Exception):
        """
        installer dep_and_conflict_error event
        will run when package has conflict/dependency error while installing it
        `ex` argument can be DependencyError or ConflictError
        """
        pr.e(
            ansi.red +\
            'Error while installing ' + pkg.data['name'] + ' (' + pkg.data['version'] + '):',
        )
        if isinstance(ex, DependencyError):
            pr.e('\tdependency error:')
        elif isinstance(ex, ConflictError):
            pr.e('\tconflict error:')
        pr.e('\t\t' + str(ex))
        pr.e(ansi.reset)

        return 1

    def arch_error_event(self, pkg: BaseArchive):
        """
        installer arch_error event
        will run when package architecture is not supported on the system
        for example while installing `amd64` package on `i386` system
        """
        pr.e(ansi.red + 'Architecture error while installing "' + pkg.data['name'] + '": your system does not support "' + pkg.data['arch'] + '" packages.' + ansi.reset)
        return 1

    def install_once(self, pkg: BaseArchive):
        """
        installs once package
        is called from install sub command
        """
        target_path = self.option_value('--target')
        if target_path == None:
            target_path = ''
        installer = Installer()

        try:
            out = installer.install(pkg,
                {
                    'cannot_read_file': self.cannot_read_file_event,
                    'invalid_json_data': self.invalid_json_data_event,
                },
                {
                    'package_currently_installed': self.package_currently_installed_event,
                    'package_new_installs': self.package_new_installs_event,
                    'package_installed': self.package_installed_event,
                    'directory_not_empty': self.directory_not_empty_event,
                    'dep_and_conflict_error': self.dep_and_conflict_error_event,
                    'arch_error': self.arch_error_event,
                },
                (not self.has_option('--auto')),
                run_scripts=(not self.has_option('--without-scripts')),
                target_path=str(target_path),
                keep_conffiles=self.has_option('--keep-conffiles'),
                check_security_blacklist=(not self.has_option('--force') and not self.has_option('-f'))
            )

            if type(out) == int:
                return out
        except PackageScriptError as ex:
            pr.e(ansi.red + 'cannot install "' + pkg.data['name'] + ':' + pkg.data['version'] + '": ' + str(ex) + ansi.reset)
            return 1
        except PackageIsInSecurityBlacklist as ex:
            pr.e(ansi.red + 'cannot install ' + pkg.data['name'] + ':' + pkg.data['version'] + ' because this is in security blacklist:')
            pr.e('  ' + ex.blacklist_item['title'] + ':')
            for l in ex.blacklist_item['description'].split('\n'):
                pr.e('\t' + l)
            pr.p(ansi.reset, end='')
            return 1
        except CannotReadFileException as ex:
            self.message(ansi.red + str(ex), True, before=ansi.reset)
            return 1
        except FileConflictError as ex:
            pr.e(ansi.red + '\nCannot install ' + pkg.data['name'] + ':' + pkg.data['version'] + ': File conflict error:')
            pr.e('    ' + str(ex))
            pr.p(ansi.reset, end='')
            return 1

    def start_run_any_script_event(self, package_name: str):
        """ will run when starting running an `any` script """
        pr.p('Processing scripts for ' + package_name + '...')

    def sub_install(self):
        """ install sub command (cati pkg install) """
        if len(self.arguments) <= 1:
            self.message('argument package file(s) required')
            return 1

        require_root_permission()

        # check transactions state before run new transactions
        pr.p('Checking transactions state...')
        state_list = BaseTransaction.state_list() # get list of undoned transactions
        if state_list:
            # the list is not empty
            StateContentShower.show(state_list)
            return 1

        packages_to_install = []
        i = 1
        while i < len(self.arguments):
            try:
                pkg = archive_factory(self.arguments[i], 'r')
                pkg.read()
                pkg.package_file_path = os.path.abspath(self.arguments[i])
                packages_to_install.append(pkg)
            except FileNotFoundError as ex:
                self.message('file "' + self.arguments[i] + '" not found' + ansi.reset, before=ansi.red)
                return 1
            except:
                self.message('cannot open "' + self.arguments[i] + '": file is corrupt' + ansi.reset, before=ansi.red)
                return 1

            i += 1

        # add packages to state
        state_f = open(Env.state_file(), 'w')
        tmp = ''
        for pkg in packages_to_install:
            tmp += ('install%' + pkg.data['name'] + '%' + pkg.data['version'] + '%' + pkg.data['arch'] + '%' + pkg.package_file_path + '\n')
        state_f.write(tmp)
        state_f.close()

        packages_to_install_names_and_versions = [pkg.data['name'] + '@' + pkg.data['version'] for pkg in packages_to_install]
        i = 0
        while i < len(packages_to_install):
            try:
                pkg = packages_to_install[i]
                tmp = self.install_once(pkg)
                if type(tmp) == int and tmp != 0:
                    if not self.has_option('--dont-ignore-state'):
                        BaseTransaction.finish_all_state()
                    return tmp
                pkg.close()
            except:
                if not self.has_option('--dont-ignore-state'):
                    BaseTransaction.finish_all_state()
                self.message('cannot install "' + packages_to_install[i].data['name'] + '"' + ansi.reset, before=ansi.red)
                return 1
            i += 1
        BaseTransaction.run_any_scripts(['install', packages_to_install_names_and_versions], events={
            'start_run_script': self.start_run_any_script_event,
        })
        BaseTransaction.finish_all_state()

    def run(self):
        """ Run command """

        if len(self.arguments) <= 0:
            pr.p(self.help_full())
            return 0
        
        if self.arguments[0] == 'build':
            return self.sub_build()
        elif self.arguments[0] == 'show':
            return self.sub_show()
        elif self.arguments[0] == 'install':
            return self.sub_install()
        else:
            self.message('unknow subcommand "' + self.arguments[0] + '"', True)
            return 1



#
# StateContentShower.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Shows and renders list of undoned transactions in state """

import sys
from cati.cmdline import pr, ansi
from cati.transaction.BaseTransaction import BaseTransaction

def show(state_list: list):
    """
    Shows and renders list of undoned transactions in state

    Args:
        state_list: loaded state from `transaction.BaseTransaction.state_list()` as (list)
    """
    pr.e(ansi.yellow + 'Error: state is not done')
    pr.p('\tthere is some undoned transactions:')
    for item in state_list:
        pr.p('\t\t' + BaseTransaction.state_item_to_string(item))
    pr.p('to complete them, run `' + sys.argv[0]  + ' state --complete`')
    pr.p(ansi.reset, end='')



#
# SecurityBlacklist.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

"""
Cati has a feature to block malware packages (detected)
this feature is a database from sha256, md5 and sha512 hashes
this database declares the malware packages.
cati checks packages hashes and checks them in this database.
if package is in database, will not install and user will get error
"this package is a malware" and details of database
"""

import os
import json
from . import Env

def get_list() -> list:
    """
    returns list of blaclist items
    
    Returns:
        list: returns list of blacklist items
    """
    db_parts = os.listdir(Env.security_blacklist())
    blacklist = []
    for part in db_parts:
        try:
            f = open(Env.security_blacklist('/' + part), 'r')
            content = f.read()
            content = json.loads(content)
            f.close()
        except:
            continue
        for item in content:
            try:
                assert type(item['title']) == str
                assert type(item['description']) == str
                assert type(item['md5']) == str
                assert type(item['sha256']) == str
                assert type(item['sha512']) == str
                blacklist.append(item)
            except:
                pass
    return blacklist



#
# DirCrawler.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Crawls an directory and extracts packages data from that """

import os
import json

class DirCrawler:
    """ Crawls an directory and extracts packages data from that """
    def __init__(self, dirpath: str):
        self.dirpath = dirpath
        self.loaded_packages = []

    def add_once(self, path: str):
        """
        add once package data item
        
        Args:
            path (str): package filepath
        """
        if path.split('.')[-1] != 'json':
            return
        try:
            f = open(path, 'r')
            content = f.read()
            f.close()
            json_data = json.loads(content.strip())
        except:
            return
        # add item
        self.loaded_packages.append(json_data)

    def start(self, path=''):
        """
        start crwaling (loaded packages will put in self.loaded_packages)
        
        Args:
            path (str): directory path
        """
        if path == '':
            path = self.dirpath
        for item in os.listdir(path):
            if os.path.isdir(path + '/' + item):
                self.start(path + '/' + item)
            else:
                self.add_once(path + '/' + item)



#
# UpdateCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Update command """

import os
import time
import random
import json
from cati.cmdline.BaseCommand import BaseCommand
from cati.cmdline import pr, ansi
from cati.cmdline.components import ReposListErrorShower, DownloadProgress
from cati.repo.Repo import Repo
from cati.frontend import Env, RootRequired
from cati.dotcati.PackageJsonValidator import PackageJsonValidator
from cati.dotcati import ListUpdater
from cati.dotcati.Pkg import Pkg

class UpdateCommand(BaseCommand):
    """ Update command """
    def help(self):
        """
        updates list of available packages from repositories

        Usage: cati update [options]

        Options:
        -q|--quiet: quiet output
        -v|--verbose: verbose output
        """
        pass

    def config(self) -> dict:
        """ Define and config this command """
        return {
            'name': 'update',
            'options': {
                '--quiet': [False, False],
                '-q': [False, False],
                '--verbose': [False, False],
                '-v': [False, False],
            },
            'max_args_count': 0,
            'min_args_count': 0,
        }

    def empty_method(self, a=None, b=None):
        """ an empty method """
        pass

    def download_event(self, url, output):
        """ repo data download event """
        return DownloadProgress.download(url, output)

    def run(self):
        """ Run command """

        RootRequired.require_root_permission()

        if not self.is_quiet():
            pr.p('Loading repositories list...')
        repos = Repo.get_list()
        ReposListErrorShower.show(repos)

        if not self.is_quiet():
            pr.p('Prepairing to update repos...')
        orig_repos = []
        for repo in repos:
            if repo.successful_loaded:
                if repo.test():
                    if not repo.is_disable:
                        orig_repos.append(repo)
                    else:
                        if not self.is_quiet():
                            self.message('Warning: ignoring repository "' + repo.name + '" because this is disable')
                else:
                    pr.e(ansi.red + 'Cannot make connection to repo "' + repo.full_string + '"' + ansi.reset)

        if not self.is_quiet():
            pr.p('Updating repositories...')
            pr.p('=============================')

        # downloaded repos data files paths
        downloaded_paths = []

        # update repos
        for repo in list(reversed(orig_repos)):
            if not self.is_quiet():
                pr.p('Fetching ' + repo.name + ' (' + repo.url + ') data...')
            data = repo.get_data(download_event=self.download_event)
            if type(data) == int:
                pr.e(ansi.red + 'Cannot update ' + repo.name + ' (' + repo.url + '): error code ' + str(data) + ansi.reset)
            elif isinstance(data, Exception):
                pr.e(ansi.red + 'Cannot update ' + repo.name + ' (' + repo.url + '): ' + str(data) + ansi.reset)
            else:
                # validate data
                try:
                    tmp = json.loads(data)
                    # save data in an file
                    path = Env.cache_dir('/' + repo.name + '-' + str(time.time()) + str(random.random())) + '.json'
                    f = open(path, 'w')
                    f.write(data)
                    f.close()
                    downloaded_paths.append(path)
                except:
                    pr.e(ansi.red + 'Cannot update ' + repo.name + ' (' + repo.url + '): invalid json data recived' + ansi.reset)

        if not self.is_quiet():
            pr.p('Updating packages list...')

        # load downloaded data
        packages = []
        for path in downloaded_paths:
            f = open(path, 'r')
            data = f.read().strip()
            f.close()
            items = json.loads(data)
            for item in items:
                if PackageJsonValidator.validate(item):
                    packages.append(item)
                else:
                    pass

        for pkg in packages:
            if PackageJsonValidator.validate(pkg):
                if self.is_verbose():
                    pr.p('adding ' + pkg['name'] + ':' + pkg['version'] + ':' + pkg['arch'] + '...')
                # write package on list
                if not os.path.isdir(Env.packages_lists('/' + pkg['name'])):
                    os.mkdir(Env.packages_lists('/' + pkg['name']))
                try:
                    f = open(Env.packages_lists('/' + pkg['name'] + '/' + pkg['version'] + '-' + pkg['arch']), 'w')
                    f.write(json.dumps(pkg))
                    f.close()
                    ListUpdater.index_reverse_depends_and_conflicts(Pkg(pkg))
                except:
                    pr.e(ansi.red + 'error while adding ' + pkg['name'] + ':' + pkg['version'] + ':' + pkg['arch'] + ansi.reset)
            else:
                if self.is_verbose():
                    pr.p(ansi.yellow + 'invalid json data in an item. ignored...' + ansi.reset)

        if self.is_quiet():
            pr.p('Finishing update...')
        ListUpdater.update_indexes({
            'cannot_read_file': self.empty_method,
            'invalid_json_data': self.empty_method,
        })

        pr.p(ansi.green + 'Done.' + ansi.reset)



#
# BaseTransaction.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

"""
Transaction base model.

transactions are install/remove/upgrade/downgrade operations.
"""

import os
from cati.dotcati.Pkg import Pkg
from cati.frontend import Env
from .Calculator import Calculator

class BaseTransaction:
    """ Transaction base model """

    @staticmethod
    def finish_all_state():
        """ clear all of states """
        f = open(Env.state_file(), 'w')
        f.write('')
        f.close()

    @staticmethod
    def add_to_state(calc: Calculator):
        """ add new item to state """
        content = ''
        for item in calc.get_sorted_list():
            content += item['action'] + '%' + item['pkg'].data['name'] + '%' + item['pkg'].data['version'] + '%' + item['pkg'].data['arch'] + '\n'
        f = open(Env.state_file(), 'w')
        f.write(content)
        f.close()

    @staticmethod
    def pop_state():
        """ remove first item from state """
        f = open(Env.state_file(), 'r')
        content = f.read()
        f.close()
        content = content.strip()
        lines = content.split('\n')
        if lines:
            lines.pop(0)
        new_content = ''
        for line in lines:
            new_content += line + '\n'
        f = open(Env.state_file(), 'w')
        f.write(new_content)
        f.close()

    @staticmethod
    def state_item_to_string(state_item: dict) -> str:
        """
        Gets an dictonary as a item in state list where returned by `BaseTransaction.state_list()`
        and generates an human readable message as string to show that message to user
        """
        msg = state_item['action'] + ' ' + state_item['pkg']
        if state_item['version'] == None:
            return msg
        msg += '=' + state_item['version']
        if state_item['arch'] == None:
            return msg
        msg += '=' + state_item['arch']
        if state_item['file'] == None:
            return msg
        msg += ' (' + state_item['file'] + ')'
        return msg

    @staticmethod
    def state_list():
        """ returns list of undoned transactions from state file """
        f = open(Env.state_file(), 'r')
        content = f.read()
        f.close()
        content = content.strip().split('\n')
        content = [line.strip() for line in content]
        result = []
        for item in content:
            if item != '':
                tmp = {}
                parts = item.split('%')
                tmp['action'] = parts[0]
                tmp['pkg'] = parts[1]
                try:
                    tmp['version'] = parts[2]
                except:
                    tmp['version'] = None
                try:
                    tmp['arch'] = parts[3]
                except:
                    tmp['arch'] = None
                try:
                    tmp['file'] = parts[4]
                except:
                    tmp['file'] = None
                result.append(tmp)
        return result

    @staticmethod
    def run_any_scripts(runed_transactions: list, events: dict):
        """
        run all of `any` scripts.
        
        events:
        - start_run_script: will run when starting to run once script (gets package name)
        """
        runed_transactions_str = ''
        for rt in runed_transactions[1]:
            runed_transactions_str += rt + ' '
        runed_transactions_str = runed_transactions_str.strip()
        scripts = os.listdir(Env.any_scripts())
        for script in scripts:
            events['start_run_script'](script)
            # run script
            os.system('chmod +x "' + Env.any_scripts('/' + script) + '"')
            os.system(Env.any_scripts('/' + script) + ' ' + runed_transactions[0] + ' ' + runed_transactions_str)



#
# ListCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" List command """

from cati.cmdline.BaseCommand import BaseCommand
from cati.cmdline import pr, ansi
from cati.dotcati.Pkg import Pkg

class ListCommand(BaseCommand):
    """ List command """
    def help(self):
        """
        shows list of packages

        Usage: cati list [options]

        Options:
        --installed: show only installed packages
        --installed-manual: show only manual installed packages
        --author=[author-name or author-nameS splited by `|`]: filter packages list by author name
        --maintainer=[maintainer-name or author-nameS splited by `|`]: filter packages list by maintainer name
        --category=[category-name or category-nameS splited by `|`]: filter packages list by category
        --search=[word]: search by packages name and description
        --upgradable: show list of upgradable packages
        """
        pass

    def config(self) -> dict:
        """ Define and config this command """
        return {
            'name': 'list',
            'options': {
                '--installed': [False, False],
                '--installed-manual': [False, False],
                '--quiet': [False, False],
                '-q': [False, False],
                '--verbose': [False, False],
                '-v': [False, False],
                '--upgradable': [False, False],
                '--author': [False, True],
                '--maintainer': [False, True],
                '--category': [False, True],
                '--search': [False, True],
            },
            'max_args_count': 0,
            'min_args_count': 0,
        }

    def show_once(self, package: Pkg):
        """
        show once item in loaded packages list
        """
        if self.has_option('-q') or self.has_option('--quiet'):
            pr.p(package.data['name'])
            return
        output = ansi.green + package.data['name'] + ansi.reset + '/' +  ansi.yellow + package.data['version'] + ansi.reset
        if package.installed():
            if package.is_installed_manual(package.data['name']):
                output += '/Installed-Manual:' + ansi.blue + package.installed().strip() + ansi.reset
            else:
                output += '/Installed:' + ansi.blue + package.installed().strip() + ansi.reset
        output += '/[' + package.data['repo'] + ']'

        # if verbose output wanted, show first line of description
        if self.is_verbose():
            try:
                description_summary = package.data['description'].split('\n')[0]
                if description_summary != '':
                    output += '/ ' + ansi.header + description_summary + ansi.reset
            except:
                pass

        pr.p(output)

    def run(self):
        """ Run command """

        if not self.has_option('-q') and not self.has_option('--quiet'):
            pr.p('Loading packages list...')
            pr.p('========================')
        # load list of packages
        if self.has_option('--installed'):
            # just list installed packages
            packages = Pkg.installed_list()
        elif self.has_option('--installed-manual'):
            packages = Pkg.installed_list()
            packages_list = [tmp_pkg for tmp_pkg in packages['list'] if Pkg.is_installed_manual(tmp_pkg.data['name'])]
            packages['list'] = packages_list
        else:
            packages = Pkg.all_list()

        for error in packages['errors']:
            self.message(error + ansi.reset, True, before=ansi.red)

        # load filter options
        wanted_authors = []
        if self.has_option('--author'):
            wanted_authors = self.option_value('--author').strip().split('|')
            wanted_authors = [author.strip() for author in wanted_authors]
        wanted_maintainers = []
        if self.has_option('--maintainer'):
            wanted_maintainers = self.option_value('--maintainer').strip().split('|')
            wanted_maintainers = [maintainer.strip() for maintainer in wanted_maintainers]
        wanted_categories = []
        if self.has_option('--category'):
            wanted_categories = self.option_value('--category').strip().split('|')
            wanted_categories = [category.strip() for category in wanted_categories]
        search_query = self.option_value('--search')
        if search_query:
            search_query_words = search_query.strip().split(' ')
            search_query_words = [word.strip() for word in search_query_words]
        else:
            search_query_words = []

        for package in packages['list']:
            # check filters
            if wanted_authors:
                try:
                    if not package.data['author'].strip() in wanted_authors:
                        continue
                except:
                    continue
            if wanted_maintainers:
                try:
                    if not package.data['maintainer'].strip() in wanted_maintainers:
                        continue
                except:
                    continue
            if wanted_categories:
                try:
                    if not package.data['category'].strip() in wanted_categories:
                        continue
                except:
                    continue
            if search_query:
                is_math_with_query = False
                for word in search_query_words:
                    if word in package.data['name']:
                        is_math_with_query = True
                try:
                    if search_query in package.data['description']:
                        is_math_with_query = True
                except:
                    pass
                if not is_math_with_query:
                    continue

            if self.has_option('--upgradable'):
                if package.installed():
                    if Pkg.compare_version(package.data['version'], package.installed()) != 1:
                        continue
                else:
                    continue
            # show item
            self.show_once(package)



#
# ClearCacheCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" ClearCache command """

import os
from cati.cmdline.BaseCommand import BaseCommand
from cati.cmdline import pr, ansi
from cati.frontend import Env, RootRequired

class ClearCacheCommand(BaseCommand):
    """ ClearCache command """
    def help(self):
        """
        clears cache files

        Usage: sudo cati clear-cache [options]

        Options:
        -v|--verbose: verbose output
        """
        pass

    def config(self) -> dict:
        """ Define and config this command """
        return {
            'name': 'clear-cache',
            'options': {
                '-v': [False, False],
                '--verbose': [False, False],
            },
            'max_args_count': 0,
            'min_args_count': 0,
        }

    def run(self):
        """ Run command """

        RootRequired.require_root_permission()

        for f in os.listdir(Env.cache_dir()):
            if os.path.isfile(Env.cache_dir('/' + f)):
                if self.is_verbose():
                    pr.p('removing ' + Env.cache_dir('/' + f) + '...')
                os.remove(Env.cache_dir('/' + f))
        for f in os.listdir(Env.cache_dir('/archives')):
            if os.path.isfile(Env.cache_dir('/archives/' + f)):
                if self.is_verbose():
                    pr.p('removing ' + Env.cache_dir('/archives/' + f) + '...')
                os.remove(Env.cache_dir('/archives/' + f))

        pr.p(ansi.green + 'Cache files cleared successfully' + ansi.reset)



#
# Remove.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Remove transaction """

import os
import shutil
from cati.transaction.BaseTransaction import BaseTransaction
from cati.dotcati.Pkg import Pkg
from cati.frontend import Env

class Remove(BaseTransaction):
    """ Remove transaction """
    @staticmethod
    def add_to_unremoved_conffiles(pkg: Pkg, filepath: str):
        """ Adds filepath to list of unremoved conffiles """
        f = open(Env.unremoved_conffiles(), 'r')
        filelist = f.read().strip().split('\n')
        f.close()

        # add item to list
        if not filepath in filelist:
            filelist.append(filepath)

        # generate new content of unremoved_conffiles file
        new_content = ''
        for item in filelist:
            new_content += item + '\n'

        # write new content to file
        f = open(Env.unremoved_conffiles(), 'w')
        f.write(new_content)
        f.close()

    @staticmethod
    def run(pkg: Pkg, events: dict, remove_conffiles=False, run_scripts=True):
        """ Remove pkg """
        events['removing_package'](pkg)

        # run rm-before script
        if run_scripts:
            if os.path.isfile(Env.installed_lists('/' + pkg.data['name'] + '/rm-before')):
                os.system('chmod +x "' + Env.installed_lists('/' + pkg.data['name'] + '/rm-before') + '"')
                with_conffiles_arg = 'without-conffiles'
                if remove_conffiles:
                    with_conffiles_arg = 'with-conffiles'
                os.system(Env.installed_lists('/' + pkg.data['name'] + '/rm-before') + ' ' + with_conffiles_arg)

        # remove package
        installed_files = open(Env.installed_lists('/' + pkg.data['name'] + '/files'), 'r').read()
        installed_files = installed_files.strip().split('\n')
        for f in list(reversed(installed_files)):
            if f != '':
                f_type = f.strip().split(':', 1)[0]
                f_path = f.strip().split(':', 1)[1]
                if f_type == 'f':
                    if os.path.isfile(Env.base_path(f_path)):
                        os.remove(Env.base_path(f_path))
                elif f_type == 'd':
                    try:
                        os.rmdir(Env.base_path(f_path))
                    except:
                        events['dir_is_not_empty'](pkg, f)
                elif f_type == 'cf':
                    if remove_conffiles:
                        if os.path.isfile(Env.base_path(f_path)):
                            os.remove(Env.base_path(f_path))
                    else:
                        Remove.add_to_unremoved_conffiles(pkg, f_path)
                elif f_type == 'cd':
                    if remove_conffiles:
                        try:
                            os.rmdir(Env.base_path(f_path))
                        except:
                            events['dir_is_not_empty'](pkg, f)
                    else:
                        Remove.add_to_unremoved_conffiles(pkg, f_path)

        # run rm-after script
        if run_scripts:
            if os.path.isfile(Env.installed_lists('/' + pkg.data['name'] + '/rm-after')):
                with_conffiles_arg = 'without-conffiles'
                if remove_conffiles:
                    with_conffiles_arg = 'with-conffiles'
                os.system('chmod +x "' + Env.installed_lists('/' + pkg.data['name'] + '/rm-after') + '"')
                os.system(Env.installed_lists('/' + pkg.data['name'] + '/rm-after') + ' ' + with_conffiles_arg)

        # remove installation config
        shutil.rmtree(Env.installed_lists('/' + pkg.data['name']))

        # remove any script
        if os.path.isfile(Env.any_scripts('/' + pkg.data['name'])):
            os.remove(Env.any_scripts('/' + pkg.data['name']))

        events['package_remove_finished'](pkg)



#
# StateCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" State command """

from cati.cmdline.BaseCommand import BaseCommand
from cati.cmdline import pr, ansi
from cati.cmdline.components import StateContentShower
from .RemoveCommand import RemoveCommand
from .PkgCommand import PkgCommand
from cati.cmdline import ArgParser
from cati.transaction.BaseTransaction import BaseTransaction
from cati.frontend.RootRequired import require_root_permission
from cati.dotcati.ArchiveModel import archive_factory

class StateCommand(BaseCommand):
    """ State command """
    def help(self):
        """
        manage transactions state

        Usage: cati state [options]
        
        Options:
        --abort: cancel undoned transactions
        --complete: complete undoned transactions
        -y|--yes: do not ask for user confirmation
        """
        pass

    def config(self) -> dict:
        """ Define and config this command """
        return {
            'name': 'state',
            'options': {
                '--abort': [False, False],
                '--complete': [False, False],
                '--yes': [False, False],
                '-y': [False, False],
            },
            'max_args_count': 0,
            'min_args_count': 0,
        }

    def run(self):
        """ Run command """

        if self.has_option('--abort'):
            require_root_permission()
            state_list = BaseTransaction.state_list()
            if not state_list:
                pr.p(ansi.green + 'There is not any undoned transaction and everything is ok' + ansi.reset)
                return 0
            user_answer = 'y'
            if not self.has_option('-y') and not self.has_option('--yes'):
                pr.p(ansi.yellow + 'WARNING: this process maybe dangerous and ' + str(len(state_list)) + ' transactions will be igonred. are you sure? [y/N] ' + ansi.reset, end='')
                user_answer = input()
            if user_answer in ['Y', 'y']:
                BaseTransaction.finish_all_state()
                pr.p(ansi.green + 'state was empty successfully' + ansi.reset)
            else:
                return
        elif self.has_option('--complete'):
            require_root_permission()
            state_list = BaseTransaction.state_list()
            if not state_list:
                pr.p(ansi.green + 'There is not any undoned transaction and everything is ok' + ansi.reset)
                return 0
            BaseTransaction.finish_all_state()
            # complete transactions
            for item in state_list:
                if item['action'] == 'remove':
                    tmp_arguments = [item['pkg'], '-y']
                    tmp_arguments.insert(0, 'cati')
                    tmp_arguments.insert(1, 'remove')
                    cmd = RemoveCommand()
                    cmd.handle(ArgParser.parse(tmp_arguments))
                elif item['action'] == 'install':
                    tmp_arguments = ['install', item['file']]
                    tmp_arguments.insert(0, 'cati')
                    tmp_arguments.insert(1, 'pkg')
                    cmd = PkgCommand()
                    cmd.handle(ArgParser.parse(tmp_arguments))
            return
        else:
            # show list of undoned transactions
            state_list = BaseTransaction.state_list()
            if state_list:
                StateContentShower.show(state_list)
                return 1
            else:
                pr.p(ansi.green + 'There is not any undoned transaction and everything is ok' + ansi.reset)
                return 0



#
# DownloadCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Download command """

import os
import shutil
from cati.cmdline.BaseCommand import BaseCommand
from cati.cmdline import pr, ansi
from cati.cmdline.components import DownloadProgress
from cati.dotcati.Pkg import Pkg

class DownloadCommand(BaseCommand):
    """ Download command """
    def help(self):
        """
        download packages

        Usage: cati download pkg1 pkg1 ... [options]
        ...... cati download pkg=<version> pkg2=<version> pkg3 ... [options]
        ...... cati download pkg=<version>=<arch> ... [options]

        Options:
        --output=[output file path]: set file download path
        -q|--quiet: quiet output
        """
        pass

    def config(self) -> dict:
        """ Define and config this command """
        return {
            'name': 'download',
            'options': {
                '-q': [False, False],
                '--quiet': [False, False],
                '--output': [False, True],
            },
            'max_args_count': None,
            'min_args_count': 1,
        }

    def download_once(self, pkg, output=None):
        """ Download once package """

        try:
            file_path = pkg.data['file_path']
        except:
            self.message('package "' + pkg.data['name'] + '" is local and cannot be downloaded' + ansi.reset, is_error=True, before=ansi.red)
            return False
        if not self.is_quiet():
            pr.p('Downloading ' + pkg.data['name'] + ':' + pkg.data['version'] + ':' + pkg.data['arch'] + '...')

        if output == None:
            output = file_path.split('/')[-1]

        if file_path[:7] == 'http://' or file_path[:8] == 'https://':
            i = 0
            while i < 5:
                res = DownloadProgress.download(file_path, output)
                if res == True:
                    break
                else:
                    pr.e(ansi.red + str(res) + ansi.reset)
                i += 1
            if i == 5:
                return False
        else:
            if not os.path.isfile(file_path):
                return False
            shutil.copy(file_path, output)

        return True

    def run(self):
        """ Run command """

        if not self.is_quiet():
            pr.p('Loading packages list...')
            pr.p('========================')

        loaded_packages = []

        for argument in self.arguments:
            arg_parts = argument.split('=')
            if len(arg_parts) == 1:
                # load last version as default
                pkg = Pkg.load_last(argument)
            elif len(arg_parts) == 2:
                # load specify version
                pkg = Pkg.load_version(arg_parts[0], arg_parts[1])
                if pkg == 1:
                    pkg = False
                elif pkg == 2:
                    self.message('package "' + arg_parts[0] + '" has not version "' + arg_parts[1] + '"' + ansi.reset, before=ansi.red)
                    continue
            else:
                # load specify version and specify arch
                pkg = Pkg.load_version(arg_parts[0], arg_parts[1], arg_parts[2])
                if pkg == 1:
                    pkg = False
                elif pkg == 2:
                    self.message('package "' + arg_parts[0] + '" has not version or arch "' + arg_parts[1] + ':' + arg_parts[2] + '"' + ansi.reset, before=ansi.red)
                    continue
            if pkg:
                loaded_packages.append(pkg)
            else:
                self.message('unknow package "' + argument + '"' + ansi.reset, before=ansi.red)

        if not loaded_packages:
            return 1

        # download loaded packages
        is_any_success = False
        output_path = self.option_value('--output')
        for pkg in loaded_packages:
            if len(loaded_packages) > 1:
                tmp = self.download_once(pkg)
            else:
                tmp = self.download_once(pkg, output_path)
            if tmp:
                is_any_success = True

        if not is_any_success:
            return 1



#
# RemoveCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Remove command """

from cati.cmdline.BaseCommand import BaseCommand
from cati.cmdline import pr, ansi
from cati.cmdline.components import TransactionShower, StateContentShower
from cati.dotcati.Pkg import Pkg
from cati.transaction.Calculator import Calculator
from cati.transaction.Remove import Remove
from cati.transaction.BaseTransaction import BaseTransaction
from cati.frontend.RootRequired import require_root_permission

class RemoveCommand(BaseCommand):
    """ Remove command """
    def help(self):
        """
        remove packages

        Usage: cati remove pkg1 pkg2 [options]

        Options:
        -y|--yes: do not ask for user confirmation
        --conffiles: also remove conffiles (full remove)
        --without-scripts: do not run package scripts in remove process
        --force|-f: force remove essential packages
        """
        pass

    def config(self) -> dict:
        """ Define and config this command """
        return {
            'name': 'remove',
            'options': {
                '-y': [False, False],
                '--yes': [False, False],
                '--conffiles': [False, False],
                '--without-scripts': [False, False],
                '--force': [False, False],
                '-f': [False, False],
            },
            'max_args_count': None,
            'min_args_count': 1,
        }

    def removing_package_event(self, pkg: Pkg):
        """
        will run as package remover event while starting removing a package
        """
        pr.p(
            'Removing ' + ansi.yellow + pkg.data['name'] + ' (' + pkg.data['version'] + ')' + ansi.reset + '...',
            end=' '
        )

    def package_remove_finished_event(self, pkg: Pkg):
        """
        will run as package remover event when package remove process finished
        """
        pr.p(ansi.green + 'OK' + ansi.reset)

    def dir_is_not_empty_event(self, pkg: Pkg, f: str):
        """
        will run as package remover event when remover wants to remove a directory
        but that dir is not empty. this event shows a warning to user
        """
        self.message('warning: directory "' + f.split(':', 1)[1] + '" is not emptry and will not be delete' + ansi.yellow, before=ansi.reset)

    def start_run_any_script_event(self, package_name: str):
        """ will run when starting running an `any` script """
        pr.p('Processing scripts for ' + package_name + '...')

    def run(self):
        """ Run command """

        require_root_permission()

        # check transactions state before run new transactions
        pr.p('Checking transactions state...')
        state_list = BaseTransaction.state_list() # get list of undoned transactions
        if state_list:
            # the list is not empty
            StateContentShower.show(state_list)
            return 1

        pr.p('Loading packages list...')
        pr.p('==============================')
        # load list of packages
        packages = []
        for arg in self.arguments:
            pkg = Pkg.load_last(arg)
            if pkg == False:
                self.message('unknow package "' + arg + '"' + ansi.reset, before=ansi.red)
            else:
                if pkg.installed():
                    packages.append(pkg)
                else:
                    self.message('package "' + pkg.data['name'] + '" is not installed' + ansi.reset, before=ansi.red)

        # start removing loaded packages
        calc = Calculator()
        calc.remove(packages)

        # show transactions
        TransactionShower.show(calc)

        essential_packages = []
        for pkg in calc.to_remove:
            try:
                if pkg.data['essential']:
                    essential_packages.append(pkg)
            except:
                pass

        if not self.has_option('--force') and not self.has_option('-f'):
            for pkg in essential_packages:
                pr.p(ansi.red + 'Package "' + pkg.data['name'] + '" is a essential package and cannot be remove. use --force|-f option to force remove them' + ansi.reset)
            if essential_packages:
                return 1

        if not calc.has_any_thing():
            return

        # check user confirmation
        if not self.has_option('-y') and not self.has_option('--yes'):
            pr.p('Do you want to continue? [Y/n] ', end='')
            answer = input()
            if not (answer == 'y' or answer == 'Y' or answer == ''):
                pr.p(ansi.yellow + 'Abort.' + ansi.reset)
                pr.exit(1)

        # add packages to state
        BaseTransaction.add_to_state(calc)

        packages_to_remove_names_and_versions = [pkg.data['name'] + '@' + pkg.data['version'] for pkg in calc.to_remove]

        # run transactions
        for pkg in calc.to_remove:
            Remove.run(
                pkg,
                {
                    'removing_package': self.removing_package_event,
                    'package_remove_finished': self.package_remove_finished_event,
                    'dir_is_not_empty': self.dir_is_not_empty_event,
                },
                self.has_option('--conffiles'),
                run_scripts=(not self.has_option('--without-scripts'))
            )
            BaseTransaction.pop_state()

        BaseTransaction.run_any_scripts(['remove', packages_to_remove_names_and_versions], events={
            'start_run_script': self.start_run_any_script_event,
        })

        BaseTransaction.finish_all_state()



#
# FinfoCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Finfo command """

import os
from cati.cmdline.BaseCommand import BaseCommand
from cati.cmdline import pr, ansi

class FinfoCommand(BaseCommand):
    """ Finfo command """
    def help(self):
        """
        shows info about an file

        this command says an file is for which package and some another details

        Usage: cati finfo /path/to/file
        """
        pass

    def config(self) -> dict:
        """ Define and config this command """
        return {
            'name': 'finfo',
            'options': {
            },
            'max_args_count': 1,
            'min_args_count': 1,
        }

    def run(self):
        """ Run command """

        if not os.path.exists(self.arguments[0]):
            pr.e('file "' + self.arguments[0] + '" not exists.')
            return 1

        filepath = os.path.abspath(self.arguments[0])

        return os.system(self.cati_exec + ' files --installed | grep ": ' + filepath + '"')



#
# FilesCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Files command """

from cati.cmdline.BaseCommand import BaseCommand
from cati.cmdline import pr, ansi
from cati.dotcati.Pkg import Pkg

class FilesCommand(BaseCommand):
    """ Files command """
    def help(self):
        """
        shows files list of packages

        Usage: cati files pkg1 pkg2 ...

        Options:
        --installed: shows list of all of installed files/dirs
        """
        pass

    def config(self) -> dict:
        """ Define and config this command """
        return {
            'name': 'files',
            'options': {
                '--installed': [False, False],
                '--quiet': [False, False],
                '-q': [False, False],
            },
            'max_args_count': None,
            'min_args_count': None,
        }

    def run(self):
        """ Run command """

        if self.has_option('--installed'):
            if not self.is_quiet():
                pr.p('Loading files list...')
                pr.p('=====================')
            all_of_installed_files_list = Pkg.get_all_installed_files_list()
            for item in all_of_installed_files_list:
                if self.is_quiet():
                    pr.p(item[2])
                else:
                    message = ''
                    if item[1] == 'd':
                        message = ' (directory)'
                    elif item[1] == 'cf':
                        message = ' (config file)'
                    elif item[1] == 'cd':
                        message = ' (config directory)'
                    pr.p(item[0] + ': ' + item[2] + message)
            return 0

        if not self.arguments:
            self.message('argument package names is required', True)
            return 1

        if not self.is_quiet():
            pr.p('Loading packages list...')
            pr.p('========================')

        loaded_packages = []

        for argument in self.arguments:
            arg_parts = argument.split('=')
            if len(arg_parts) == 1:
                # load last version as default
                pkg = Pkg.load_last(argument)
            else:
                # load specify version
                pkg = Pkg.load_version(arg_parts[0], arg_parts[1])
                if pkg == 1:
                    pkg = False
                elif pkg == 2:
                    self.message('package "' + arg_parts[0] + '" has not version "' + arg_parts[1] + '"' + ansi.reset, before=ansi.red)
                    continue
            if pkg:
                loaded_packages.append(pkg)
            else:
                self.message('unknow package "' + argument + '"' + ansi.reset, before=ansi.red)

        if not loaded_packages:
            return 1

        # show files list of loaded packages
        for pkg in loaded_packages:
            try:
                files_list = pkg.data['files']
            except:
                files_list = []
            if not self.is_quiet():
                pr.p(pkg.data['name'] + ':')
                if not files_list:
                    pr.p(ansi.yellow + '  This package is empty' + ansi.reset)
                for item in files_list:
                    pr.p('  ' + item)
            else:
                for item in files_list:
                    pr.p(item)
            if len(loaded_packages) > 1:
                if not self.is_quiet():
                    pr.p('========================')



#
# ListUpdater.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Database Packages list updater """

import os
import json
from cati.frontend.RootRequired import require_root_permission
from cati.frontend import Env
from .Pkg import Pkg

def update_indexes(events: dict):
    """
    This function loads available versions of a package and index them in index file
    and do this action for all of packages in lists.

    Args:
        events: (dict) the `events` argument should be a dictonary from functions. this will use to handle errors
                for example if some thing went wrong, the spesific function in events
                will run.
                events:
                - cannot_read_file: if in this process, an error happened while reading a file, this will run with file path arg
                - invalid_json_data: if json content of a file is curropt, this will run with file path and content args
    """

    require_root_permission()

    for pkg in os.listdir(Env.packages_lists()):
        pkg_index = {}
        if os.path.isdir(Env.packages_lists('/' + pkg)):
            for version in os.listdir(Env.packages_lists('/' + pkg)):
                if version not in ['index', 'reverse_depends', 'reverse_conflicts']:
                    if os.path.isfile(Env.packages_lists('/' + pkg + '/' + version)):
                        content = None
                        try:
                            f = open(Env.packages_lists('/' + pkg + '/' + version), 'r')
                            content = f.read()
                        except:
                            events['cannot_read_file'](Env.packages_lists('/' + pkg + '/' + version))

                        if content != None:
                            try:
                                content_json = json.loads(content)
                                try:
                                    tmp = pkg_index[content_json['arch']]
                                    del tmp
                                except:
                                    pkg_index[content_json['arch']] = []
                                pkg_index[content_json['arch']].append(content_json['version'])
                            except:
                                events['invalid_json_data'](Env.packages_lists('/' + pkg + '/' + version), content)
        # write generated index to index file
        f_index = open(Env.packages_lists('/' + pkg + '/index'), 'w')
        f_index.write(json.dumps(pkg_index))
        f_index.close()

def index_reverse_depends_and_conflicts(pkg: Pkg):
    """
    Packages have `depends` and `conflicts`
    But also they have `Reverse` depends and conflicts
    Reverse d/c should be indexed because loading them real time is so big process
    We index them in a place, and when a package is added/changed, this function should be called

    Args:
        pkg (Pkg): changed/added package (reverse c/d will be set for that packages this package is related to them)
    """
    # load the packages
    depend_pkgs = []
    conflict_pkgs = []
    for depend in pkg.get_depends():
        query_parts = Pkg.check_state(depend, only_parse=True)
        for depth1 in query_parts:
            for depth2 in depth1:
                depend_pkgs.append(depth2[0])
    for conflict in pkg.get_conflicts():
        query_parts = Pkg.check_state(conflict, only_parse=True)
        for depth1 in query_parts:
            for depth2 in depth1:
                conflict_pkgs.append(depth2[0])

    # set reverse depends/conflicts for found packages
    for p in depend_pkgs:
        f_path = Env.packages_lists('/' + p + '/reverse_depends')
        current_list = None
        try:
            if not os.path.isdir(Env.packages_lists('/' + p)):
                os.mkdir(Env.packages_lists('/' + p))
            if not os.path.isfile(f_path):
                current_list = []
            else:
                f = open(f_path, 'r')
                current_list = [item.strip() for item in f.read().strip().split('\n') if item.strip() != '']
                f.close()
            if not pkg.data['name'] in current_list:
                current_list.append(pkg.data['name'])
            # write new list
            new_list_str = ''
            for item in current_list:
                new_list_str += item + '\n'
            new_list_str = new_list_str.strip()
            f = open(f_path, 'w')
            f.write(new_list_str)
            f.close()
        except:
            pass
    for p in conflict_pkgs:
        f_path = Env.packages_lists('/' + p + '/reverse_conflicts')
        current_list = None
        try:
            if not os.path.isdir(Env.packages_lists('/' + p)):
                os.mkdir(Env.packages_lists('/' + p))
            if not os.path.isfile(f_path):
                current_list = []
            else:
                f = open(f_path, 'r')
                current_list = [item.strip() for item in f.read().strip().split('\n') if item.strip() != '']
                f.close()
            if not pkg.data['name'] in current_list:
                current_list.append(pkg.data['name'])
            # write new list
            new_list_str = ''
            for item in current_list:
                new_list_str += item + '\n'
            new_list_str = new_list_str.strip()
            f = open(f_path, 'w')
            f.write(new_list_str)
            f.close()
        except:
            pass



#
# MeowCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Meow command """

from cati.cmdline.BaseCommand import BaseCommand
from cati.cmdline import pr

class MeowCommand(BaseCommand):
    """ Meow command """
    def help(self):
        """
        Cati is a super cat, Meowwwww...!
        """
        pass

    def config(self) -> dict:
        """ Define and config this command """
        return {
            'name': 'meow',
            'options': {
            },
            'max_args_count': None,
            'min_args_count': None,
        }

    def run(self):
        """ Run command """
        pr.p("""
               /\       /\            * * * * * * * * * * * * * *
               | |_____| |            * Meow.....! I am Cati!   *
              |  --   --  |       ....* I can manage your       *
             |  { { * } }  | ..../    * Packages!               *
              |           |           *                         *
     __________|         |            * * * * * * * * * * * * * *
    /                     |
   /                      |
__/_  __  _________  _   _|
    |_| |_|       |_| |_|
""")



#
# Repo.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Cati repository model """

import os
import json
from cati.frontend import SysArch, Env
from .drivers.File import File as FileDriver
from .drivers.Http import Http as HttpDriver

class Repo:
    """ Cati repository model """
    def __init__(self, config: str):
        # ignore comments
        self.full_string = config.strip()
        config = config.strip().split(' ')
        self.url = config[0]
        self.syntax_errors = []
        self.successful_loaded = True
        config.pop(0)

        # set default values
        self.pkg = ['cati']
        self.arch = [SysArch.sys_arch()]
        self.channel = ['default']
        self.priority = 1
        self.is_disable = False
        self.name = None

        for item in config:
            item = item.strip()
            if item != '':
                parts = item.split('=')
                if len(parts) <= 1:
                    self.syntax_errors.append('undefined value for `' + parts[0] + '`')
                else:
                    if parts[0] == 'pkg':
                        self.pkg = parts[1].strip().split(',')
                        self.pkg = [tmp.strip() for tmp in self.pkg if tmp.strip() != '']
                    elif parts[0] == 'arch':
                        self.arch = parts[1].strip().split(',')
                        self.arch = [tmp.strip() for tmp in self.arch if tmp.strip() != '']
                    elif parts[0] == 'channel':
                        self.channel = parts[1].strip().split(',')
                        self.channel = [tmp.strip() for tmp in self.channel if tmp.strip() != '']
                    elif parts[0] == 'priority':
                        try:
                            self.priority = int(parts[1])
                        except:
                            pass
                    elif parts[0] == 'name':
                        self.name = parts[1]
                    elif parts[0] == 'disable':
                            self.is_disable = True
                    else:
                        self.syntax_errors.append('unknow option `' + parts[0] + '`')

        if self.name == None:
            self.name = self.pkg[0] + '-' + self.arch[0] + '-' + self.channel[0] + '-' + str(self.priority)
        if self.name == 'Local' or self.name == 'local':
            self.name = self.name + '-repo'

        # initialize driver by url
        url_parts = self.url.strip().split('://', 1)
        if len(url_parts) <= 1:
            self.successful_loaded = False
            self.syntax_errors.append('invalid url')
            return
        # find driver
        driver_name = url_parts[0]
        if driver_name == 'file':
            self.__driver = FileDriver(url=self.url, pkg=self.pkg, arch=self.arch)
        elif driver_name in ['http', 'https']:
            self.__driver = HttpDriver(url=self.url, pkg=self.pkg, arch=self.arch)
        else:
            self.successful_loaded = False
            self.syntax_errors.append('invalid url type')
            return

    def test(self) -> bool:
        """
        Test repository
        
        Returns:
            bool: True means connection is ok, False means not
        """
        return self.__driver.test()

    def get_data(self, download_event=None) -> str:
        """
        Recives repo data returns data as json

        Args:
            download_event (callable, None): if repo driver is http, this should be a function to download data

        Returns:
            str: repository data
        """
        data = self.__driver.get_data(download_event)
        try:
            j = json.loads(data)
            i = 0
            while i < len(j):
                try:
                    if not j[i]['arch'] in self.arch or not j[i]['pkg_type'] in self.pkg:
                        j.pop(i)
                except:
                    j.pop(i)
                i += 1
            data = json.dumps(j)
        except:
            pass
        data = json.loads(data)
        i = 0
        while i < len(data):
            try:
                tmp_channel = data[i]['channel']
                if self.channel != ['default']:
                    if not tmp_channel in self.channel:
                        data.pop(i)
                        continue
            except:
                pass
            data[i]['repo'] = self.name
            try:
                data[i]['file_size']
            except:
                data[i]['file_size'] = 0
            i += 1
        data = json.dumps(data)
        return data

    def get_pkg_str(self) -> str:
        """
        Returns self.pkg list (allowed package type of repo) as string

        Returns:
            str
        """
        s = ''
        for pkg in self.pkg:
            s += pkg + ','
        return s.strip(',')

    def get_arch_str(self) -> str:
        """
        Returns self.arch list (allowed package arch of repo) as string

        Returns:
            str
        """
        s = ''
        for arch in self.arch:
            s += arch + ','
        return s.strip(',')

    def get_channel_str(self) -> str:
        """
        Returns self.channel list (allowed version channels of repo) as string

        Returns:
            str
        """
        s = ''
        for channel in self.channel:
            s += channel + ','
        return s.strip(',')

    @staticmethod
    def get_list() -> list:
        """
        returns list of repositories
        
        Returns:
            list[Repo]: list of loaded repositories
        """
        repos = []
        files = [Env.repos_config()]
        for item in os.listdir(Env.repos_config_dir()):
            if os.path.isfile(Env.repos_config_dir('/' + item)):
                files.append(Env.repos_config_dir('/' + item))
        for fl in files:
            f = open(fl, 'r')
            repos_content = f.read()
            f.close()
            lines = repos_content.split('\n')
            line_counter = 1
            for line in lines:
                line = line.split('#')[0].strip()
                line = line.strip()
                if line != '':
                    repo = Repo(line)
                    repo.line_number = line_counter
                    repo.loaded_from_file = fl
                    repos.append(repo)
                line_counter += 1

        # sort by priority
        sorted_repos = []
        while len(repos):
            i = 0
            while i < len(repos):
                is_less_than_all = True
                j = 0
                while j < len(repos):
                    if int(repos[i].priority) > int(repos[j].priority):
                        is_less_than_all = False
                    j += 1
                if is_less_than_all:
                    sorted_repos.append(repos[i])
                    repos.pop(i)
                i += 1

        return sorted_repos



#
# __init__.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

import sys
from cati.cmdline import kernel, pr, ansi
from cati.frontend import HealthChecker
from cati.frontend.Version import version as __version__

# check cati installation health
def cati_installation_is_corrupt(filepath: str, filetype: str):
    """
    Will run when cati installation is corrupt
    shows error to user
    """
    pr.e(ansi.red + 'Cati installation is corrupt. to repair it, just run cati with root access' + ansi.reset)
    pr.exit(1)

def run():
    # check the installation health
    HealthChecker.check({
        'failed_to_repair': cati_installation_is_corrupt,
    })
    # handle cli
    kernel.handle(sys.argv[:])



#
# ArgParser.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Cli argument parser """

from cati.cmdline import ansi

def parse(args: list) -> dict:
    """
    Gets a list from program arguments and returns parsed args

    Args:
        args: (list) command line arguments to parse that

    Returns:
        (dict) parsed arguments output structure:
        {
            ## all of arguments where starts with `-` (dict)
            "options": {
                "--option1": "value",
                "--option2": None, ## value will be None when option has not value
                ## ...
            },

            ## all of non-option arguemnt (list)
            "arguments": [
                "arg1",
                "arg2",
                ## ...
            ]
        }
    """

    # pop first argument (cati self exec)
    args.pop(0)

    tmp_options = []
    arguments = []

    # split options & arguments
    for arg in args:
        if arg[0] == '-':
            tmp_options.append(arg)
        else:
            arguments.append(arg)

    options = {}
    for option in tmp_options:
        op_parts = option.split('=', 1)
        if len(op_parts) == 1:
            options[option] = None
        else:
            options[op_parts[0]] = op_parts[1]

    # check for --no-ansi
    i = 0
    for k in options:
        if k == '--no-ansi':
            del options[k]
            # disable ansi
            ansi.disable()
            break

    return {
        'options': options,
        'arguments': arguments
    }



#
# Pkg.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Package model """

import os
import json
from cati.frontend import Env
from cati.frontend.SysArch import sys_arch
from .exceptions import CannotReadFileException
from packaging import version
from cati.helpers.hash import calc_file_sha256

class Pkg:
    """ Package model """

    def __init__(self, data: dict):
        self.data = data
        try:
            self.data['repo']
        except:
            self.data['repo'] = 'Local'
        self.data['version'] = self.data['version'].strip()

    def installed(self) -> (str, bool):
        """ Checks current package is installed. if yes, returns installed version and if not, returns False """
        if not Pkg.is_installed(self.data['name']):
            return False

        return Pkg.installed_version(self.data['name'])

    def installed_static_files(self) -> list:
        """
        returns list of installed files of package

        Returns:
            list[str]: files paths
        """
        if not self.installed():
            return False
        installed_static_files_list = open(Env.installed_lists('/' + self.data['name'] + '/staticfiles'), 'r').read().strip()
        installed_static_files_list = installed_static_files_list.split('\n')
        installed_static_files_list = [item.strip().split('@', 1) for item in installed_static_files_list if item.strip() != '']
        return installed_static_files_list

    def get_depends(self) -> list:
        """ Returns package dependencies list """
        try:
            return self.data['depends']
        except:
            return []

    def get_recommends(self) -> list:
        """ Returns package recommends list """
        try:
            return self.data['recommends']
        except:
            return []

    def get_conflicts(self) -> list:
        """ Returns package conflicts list """
        try:
            return self.data['conflicts']
        except:
            return []

    def get_replaces(self) -> list:
        """ Returns package replaces list """
        try:
            return self.data['replaces']
        except:
            return []

    def get_conffiles(self) -> list:
        """ Returns package conffiles list """
        try:
            return self.data['conffiles']
        except:
            return []

    def get_static_files(self) -> list:
        """ returns package static files list """
        try:
            return self.data['staticfiles']
        except:
            return []

    def get_reverse_depends(self) -> list:
        """ Returns list of packages has dependency to this package
        
        Returns:
            list[Pkg]: list of packages has dependency to this package
        """
        try:
            rd_pkgs_list_f = open(Env.packages_lists('/' + self.data['name'] + '/reverse_depends'), 'r')
            rd_pkgs_list = rd_pkgs_list_f.read().strip().split('\n')
            rd_pkgs_list_f.close()
        except:
            rd_pkgs_list = []
        rd_pkgs = []
        for item in rd_pkgs_list:
            if Pkg.is_installed(item):
                rd_pkgs.append(Pkg.load_version(item, Pkg.installed_version(item)))
            else:
                rd_pkgs.append(Pkg.load_last(item))
        reverse_depends = []
        for pkg in rd_pkgs:
            if pkg.installed():
                pkg = Pkg.load_version(pkg.data['name'], pkg.installed())
            for dep in pkg.get_depends():
                if dep.strip()[0] != '@':
                    result = Pkg.check_state(dep, virtual={
                        'remove': [
                            [
                                self.data['name'], self.data['version']
                            ]
                        ],
                        'all_real_is_installed': True
                    })
                    if not result:
                        reverse_depends.append(pkg)
        return reverse_depends

    def get_reverse_conflicts(self) -> list:
        """ Returns list of packages has conflicts with this package
        
        Returns:
            list[Pkg]: list of packages has conflict with this package"""
        try:
            rd_pkgs_list_f = open(Env.packages_lists('/' + self.data['name'] + '/reverse_conflicts'), 'r')
            rd_pkgs_list = rd_pkgs_list_f.read().strip().split('\n')
            rd_pkgs_list_f.close()
        except:
            rd_pkgs_list = []
        rd_pkgs = []
        for item in rd_pkgs_list:
            if Pkg.is_installed(item):
                rd_pkgs.append(Pkg.load_version(item, Pkg.installed_version(item)))
        reverse_conflicts_list = []
        for pkg in rd_pkgs:
            conflicts = pkg.get_conflicts()
            for conflict in conflicts:
                result = Pkg.check_state(conflict, {'install': [
                    [
                        self.data['name'], self.data['version']
                    ]
                ], 'no_real_installs': True})
                if result:
                    pkg.conflict_error = conflict
                    reverse_conflicts_list.append(pkg)
        return reverse_conflicts_list

    def get_file_size_str(self) -> str:
        """
        Returns file size of the package as str (for example `32 MB`)

        Returns:
            str: file size str
        """
        try:
            file_size = self.data['file_size']
        except:
            file_size = 0

        return Pkg.get_download_size_str(file_size)

    @staticmethod
    def get_download_size_str(dsize: int) -> str:
        """
        Returns file size str from bytes interger

        Args:
            dsize: bytes count

        Returns:
            str: download str
        """
        return str(dsize) + ' B'

    def get_versions_list(self):
        """
        returns versions list of the package
        
        Returns:
            list: list of versions: [ [<version>, <arch>] ]
        """
        try:
            f_index = open(Env.packages_lists('/' + self.data['name'] + '/index'), 'r')
            index_json = json.loads(f_index.read())
            f_index.close()
        except:
            return []
        
        try:
            versions = []
            for k in index_json:
                for ver in index_json[k]:
                    versions.append([ver, k])
            return versions
        except:
            return []

    @staticmethod
    def get_all_installed_files_list() -> list:
        """
        returns list of all of installed files

        Returns:
            list: [
                ['pkgname', 'filetype(d,f,cd,cf)', '/file/path'],
                ['pkg1', 'f', '/path/to/file'],
                ...
            ]
        """
        # load list of installed packages
        installed_packages = Pkg.installed_list()['list']
        result = []
        for pkg in installed_packages:
            pkg_installed_files = open(Env.installed_lists('/' + pkg.data['name'] + '/files'), 'r').read()
            pkg_installed_files = pkg_installed_files.strip().split('\n')
            pkg_installed_files = [item.strip().split(':', 1) for item in pkg_installed_files]
            for f in pkg_installed_files:
                if len(f) > 1:
                    result.append([pkg.data['name'], f[0], f[1]])
        return result

    @staticmethod
    def load_last(pkg_name: str) -> bool:
        """ Load last version of package by name

        Args:
            pkg_name (str): the package name

        Returns:
            bool (False): package not found
            Pkg: the loaded Pkg boject
        """
        pkgs_list = Pkg.all_list()
        for item in pkgs_list['list']:
            if item.data['name'] == pkg_name:
                return item
        return False

    @staticmethod
    def is_installed(package_name: str) -> bool:
        """ Gets a package name and checks is installed or not
        
        Args:
            package_name (str): the package name you want to check is installed
        
        Returns:
            bool
        """
        try:
            assert os.path.isdir(Env.installed_lists('/' + package_name))
            assert os.path.isfile(Env.installed_lists('/' + package_name + '/ver'))
            assert os.path.isfile(Env.installed_lists('/' + package_name + '/files'))
            return True
        except:
            return False

    @staticmethod
    def is_installed_manual(package_name: str) -> bool:
        """ Gets a package name and checks is installed MANUAL or not
        
        Args:
            package_name (str): the package name you want to check is installed
        
        Returns:
            bool
        """
        if not Pkg.is_installed(package_name):
            return False
        return os.path.isfile(Env.installed_lists('/' + package_name + '/manual'))

    @staticmethod
    def load_version(pkg_name: str, version: str, arch=''):
        """
        loads a specify version of a package

        Args:
            pkg_name (str): name of package
            version (str): that version you want to load
            arch (str): load specify arch (optional)
        
        Returns:
            int (1): package not found
            int (2): package found, but version/arch not found
            Pkg object: package and version found and returned
        """
        if not os.path.isfile(Env.packages_lists('/' + pkg_name + '/index')):
            return 1

        # load package index file
        f_index = open(Env.packages_lists('/' + pkg_name + '/index'), 'r')
        index_content = f_index.read()
        f_index.close()

        try:
            # load json
            index = json.loads(index_content)
        except:
            return 1

        # load package versions list
        versions = []
        try:
            versions = index[arch]
        except:
            try:
                versions = index[SysArch.sys_arch()]
                arch = SysArch.sys_arch()
            except:
                try:
                    versions = index['all']
                    arch = 'all'
                except:
                    versions = index[list(index.keys())[0]]
                    arch = list(index.keys())[0]

        for ver in versions:
            if ver == version:
                # load this version data
                try:
                    f_ver = open(Env.packages_lists('/' + pkg_name + '/' + ver + '-' + arch), 'r')
                    f_ver_content = f_ver.read()
                    ver_data = json.loads(f_ver_content)
                    return Pkg(ver_data)
                except:
                    return 2
        return 2

    @staticmethod
    def installed_version(package_name: str) -> str:
        """ Gets name of package and returns installed version of that
        
        Args:
            package_name (str): name of package
        
        Returns:
            str: installed version
        
        Raises:
            package.exceptions.CannotReadFileException: when cannot read package database files
        """
        try:
            f = open(Env.installed_lists('/' + package_name + '/ver'), 'r')
            version = f.read()
            f.close()
        except:
            raise CannotReadFileException('cannot read file "' + Env.installed_lists('/' + package_name + '/ver') + '"')
        return version

    @staticmethod
    def installed_list() -> dict:
        """ Returns list of only installed packages 
        
        Returns:
            dict: output has two keys:
                  {
                      "list": list[Pkg] // list of packages
                      "errors": list // errors while loading packages
                  }
        """
        all_packages = Pkg.all_list()
        installed_packages = {
            'errors': all_packages['errors'],
            'list': [],
        }
        for pkg in all_packages['list']:
            if pkg.installed():
                installed_packages['list'].append(pkg)

        return installed_packages

    @staticmethod
    def all_list():
        """ Returns list of packages
        
        Returns:
            dict: output has two keys:
                  {
                      "list": list[Pkg] // list of packages
                      "errors": list // errors while loading packages
                  }
        """

        errors = []
        packages = []

        tmp_list = os.listdir(Env.packages_lists())
        tmp_list.sort()
        for item in tmp_list:
            if os.path.isfile(Env.packages_lists('/' + item + '/index')):
                f_index = open(Env.packages_lists('/' + item + '/index'), 'r')
                try:
                    index_content = f_index.read()
                    try:
                        index_json = json.loads(index_content)
                        try:
                            pkg = Pkg.load_from_index(index_json, item)
                            packages.append(pkg)
                        except:
                            if not (os.path.isfile(Env.packages_lists('/' + item + '/reverse_depends')) or os.path.isfile(Env.packages_lists('/' + item + '/reverse_conflicts'))):
                                errors.append('faild to load package "' + item + '"')
                    except:
                        errors.append('invalid json content in "' + Env.packages_lists('/' + item + '/index') + '"')
                except:
                    errors.append('cannot read file "' + Env.packages_lists('/' + item + '/index') + '"')
            else:
                errors.append(f'package "{item}" has not index file in lists ({Env.packages_lists("/" + item + "/index")} not found)')

        return {'list': packages, 'errors': errors}

    @staticmethod
    def compare_version(a: str, b: str):
        """
        Compares two versions.

        Args:
            a (str): first version
            b (str): second version
        
        Returns:
            if 1 is returned means a is upper than b.
            if 0 is returned means a equals b.
            if -1 is returned means a is less than b.
        """
        a = version.parse(a)
        b = version.parse(b)

        if a > b:
            return 1

        if a == b:
            return 0

        if a < b:
            return -1

    @staticmethod
    def get_last_version(versions: list) -> str:
        """ Gets a list from versions and returns latest version in that list
        
        Args:
            versions (list[str]): list of versions you want find last of them
        
        Returns:
            str: the last version in the list
        """
        max_ver = ''
        for version in versions:
            if Pkg.compare_version(version, max_ver) == 1:
                max_ver = version

        return max_ver

    @staticmethod
    def load_from_index(index_json: dict, package_name: str):
        """ Loads package data from index file
        
        Args:
            package_name (str): name of the package
            index_json (dict): loaded json data from `/var/lib/cati/lists/<pkgname>/index` file

        Returns:
            Pkg: the loaded Pkg object
        """
        try:
            arch = sys_arch()
            versions = index_json[arch]
        except:
            try:
                arch = 'all'
                versions = index_json[arch]
            except:
                arch = list(index_json.keys())[0]
                versions = index_json[arch]

        # load latest version
        ver = Pkg.get_last_version(versions)

        f = open(Env.packages_lists('/' + package_name + '/' + ver + '-' + arch), 'r')
        content = f.read()
        f.close()

        content_json = json.loads(content)

        return Pkg(content_json)

    def check_state(query_string: str, virtual=None, get_false_pkg=False, get_false_pkg_next=0, get_true_pkg=False, get_true_pkg_next=0, only_parse=False) -> bool:
        """
        Checks package state by query string.

        Examples:
        "somepackage >= 1.5",
        "somepackage",
        "somepackage = 2.0",
        "somepackage < 1.7",
        "pkga | pkgb >= 1.0",
        "pkga | pkgb | pkgc",
        "pkga | pkgb & pkgc = 1.0",

        also there is a feature to check files:

        "@/usr/bin/somefile",
        "somepackage | @/path/to/somefile",
        "testpkga >= 3.0 | @/somefile"
        "@/file/one | @/file/two",
        "@<sha256-hash>@/path/to/file",
        "@76883f0fd14015c93296f0e4202241f4eb3a23189dbc17990a197477f1dc441a@/path/to/file"

        `virtual` argument:

        this argument can make a virtual package state
        for example package `testpkgz` is not installed,
        but we want to check the query when this is installed
        we can set that package in virtual system to query checker
        calculate tha package as installed/removed
        virtual structure: this is dictonary:
        {
            'install': [
                ## a list from installed packages:
                ['testpkgx', '1.0'],
                ['testpkgz', '3.7.11'],
                ...
            ]

            'remove': [
                ## a list from removed packages:
                ['testpkgx', '1.0'],
                ['testpkgz', '3.7.11'],
                ...
            ]

            ## set it True if you want to ignore real installations
            'no_real_installs': True/False

            ## set it True if you want to ignore real not installations
            'all_real_is_installed': True/False
        }
        """

        # parse query string
        parts = query_string.strip().split('|')
        orig_parts = []
        for part in parts:
            tmp = part.strip().split('&')
            orig_parts.append(tmp)

        # load virtual item
        no_real_installs = False
        all_real_is_installed = False
        if virtual:
            try:
                tmp = virtual['install']
            except:
                virtual['install'] = []
            try:
                tmp = virtual['remove']
            except:
                virtual['remove'] = []
            virtual_installed_names_list = []
            virtual_installed_versions_dict = {}
            for item in virtual['install']:
                virtual_installed_versions_dict[item[0]] = item[1]
                virtual_installed_names_list.append(item[0])
            virtual_removed_names_list = []
            virtual_removed_versions_dict = {}
            for item in virtual['remove']:
                virtual_removed_versions_dict[item[0]] = item[1]
                virtual_removed_names_list.append(item[0])
            try:
                no_real_installs = virtual['no_real_installs']
            except:
                no_real_installs = False
            try:
                all_real_is_installed = virtual['all_real_is_installed']
            except:
                all_real_is_installed = False
        else:
            virtual_installed_names_list = []
            virtual_installed_versions_dict = {}
            virtual_removed_names_list = []
            virtual_removed_versions_dict = {}
            no_real_installs = False
            all_real_is_installed = False

        # parse once query
        i = 0
        while i < len(orig_parts):
            j = 0
            while j < len(orig_parts[i]):
                orig_parts[i][j] = orig_parts[i][j].strip()
                spliter = None
                if '>=' in orig_parts[i][j]:
                    spliter = '>='
                elif '<=' in orig_parts[i][j]:
                    spliter = '<='
                elif '>' in orig_parts[i][j]:
                    spliter = '>'
                elif '<' in orig_parts[i][j]:
                    spliter = '<'
                elif '=' in orig_parts[i][j]:
                    spliter = '='
                if spliter != None:
                    orig_parts[i][j] = orig_parts[i][j].split(spliter)
                    orig_parts[i][j].insert(1, spliter)
                else:
                    orig_parts[i][j] = [orig_parts[i][j]]
                z = 0
                while z < len(orig_parts[i][j]):
                    orig_parts[i][j][z] = orig_parts[i][j][z].strip()
                    z += 1
                j += 1
            i += 1

        if only_parse:
            return orig_parts

        # check query
        for tmp in orig_parts:
            ands_ok = True
            for p in tmp:
                if p[0][0] == '@':
                    # check file query
                    parts = p[0].split('@')
                    if len(parts) == 2:
                        if not os.path.exists(Env.base_path(parts[-1])):
                            ands_ok = False
                    elif len(parts) == 3:
                        if not os.path.exists(Env.base_path(parts[-1])):
                            ands_ok = False
                        else:
                            if os.path.isfile(Env.base_path(parts[-1])):
                                sha256_sum = calc_file_sha256(Env.base_path(parts[-1]))
                                if parts[1] != sha256_sum:
                                    ands_ok = False
                elif not p[0] in virtual_installed_names_list and no_real_installs:
                    ands_ok = False
                    if get_false_pkg and get_false_pkg_next <= 0:
                        return p
                    else:
                        get_false_pkg_next -= 1
                elif not p[0] in virtual_removed_names_list and all_real_is_installed:
                    pass
                elif len(p) == 1:
                    if not Pkg.is_installed(p[0]) and not p[0] in virtual_installed_names_list or p[0] in virtual_removed_names_list:
                        ands_ok = False
                        if get_false_pkg and get_false_pkg_next <= 0:
                            return p
                        else:
                            get_false_pkg_next -= 1
                elif len(p) == 3:
                    if not Pkg.is_installed(p[0]) and not p[0] in virtual_installed_names_list or p[0] in virtual_removed_names_list:
                        ands_ok = False
                        if get_false_pkg and get_false_pkg_next <= 0:
                            return p
                        else:
                            get_false_pkg_next -= 1
                    else:
                        if p[0] in virtual_installed_names_list:
                            a_ver = virtual_installed_versions_dict[p[0]]
                        else:
                            a_ver = Pkg.installed_version(p[0])
                        b_ver = p[2]
                        if p[1] == '=':
                            if Pkg.compare_version(a_ver, b_ver) != 0:
                                ands_ok = False
                                if get_false_pkg and get_false_pkg_next <= 0:
                                    return p
                                else:
                                    get_false_pkg_next -= 1
                        elif p[1] == '>':
                            if Pkg.compare_version(a_ver, b_ver) != 1:
                                ands_ok = False
                                if get_false_pkg and get_false_pkg_next <= 0:
                                    return p
                                else:
                                    get_false_pkg_next -= 1
                        elif p[1] == '<':
                            if Pkg.compare_version(a_ver, b_ver) != -1:
                                ands_ok = False
                                if get_false_pkg and get_false_pkg_next <= 0:
                                    return p
                                else:
                                    get_false_pkg_next -= 1
                        elif p[1] == '<=':
                            if Pkg.compare_version(a_ver, b_ver) != -1 and Pkg.compare_version(a_ver, b_ver) != 0:
                                ands_ok = False
                                if get_false_pkg and get_false_pkg_next <= 0:
                                    return p
                                else:
                                    get_false_pkg_next -= 1
                        elif p[1] == '>=':
                            if Pkg.compare_version(a_ver, b_ver) != 1 and Pkg.compare_version(a_ver, b_ver) != 0:
                                ands_ok = False
                                if get_false_pkg and get_false_pkg_next <= 0:
                                    return p
                                else:
                                    get_false_pkg_next -= 1
                        else:
                            ands_ok = False
                            if get_false_pkg and get_false_pkg_next <= 0:
                                return p
                            else:
                                get_false_pkg_next -= 1
                if ands_ok and get_true_pkg and get_true_pkg_next <= 0:
                    return p
                else:
                    get_true_pkg_next -= 1
            if ands_ok:
                return True

        return False



#
# UpgradeCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Upgrade command """

from cati.cmdline.BaseCommand import BaseCommand
from cati.cmdline import pr, ansi, ArgParser
from cati.dotcati.Pkg import Pkg
from .InstallCommand import InstallCommand

class UpgradeCommand(BaseCommand):
    """ Upgrade command """
    def help(self):
        """
        upgrade all of packages

        Usage: sudo cati upgrade [options]

        Options: all of `install` command options
        """
        pass

    def config(self) -> dict:
        """ Define and config this command """
        install_cmd_options = InstallCommand().config()['options']
        return {
            'name': 'upgrade',
            'options': install_cmd_options,
            'max_args_count': 0,
            'min_args_count': 0,
        }

    def run(self):
        """ Run command """

        pr.p('Checking upgradable packages...')
        installed_packages = Pkg.installed_list()['list']
        upgradable_packages = []
        for pkg in installed_packages:
            installed_version = pkg.installed()
            latest_version = pkg.data['version']
            if Pkg.compare_version(latest_version, installed_version) == 1:
                upgradable_packages.append(pkg)

        if not upgradable_packages:
            pr.p(ansi.green + 'all of packages are up to date' + ansi.reset)
            return 0

        packages_names = [pkg.data['name'] for pkg in upgradable_packages]

        install_cmd = InstallCommand()
        return install_cmd.handle(ArgParser.parse(['cati', 'install', *self.args['options'], *packages_names]))
